import abc
import datetime
import json
import os
import shutil
import sys
from typing import Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.cluster import KMeans

from beetroots.approx_optim.approach_type import utils


class ApproxParamsOptim(abc.ABC):
    r"""Optimization of an approximation parameter for the likelihood function ``beetroots.modelling.likelihoods.approx_censored_add_mult.MixingModelsLikelihood``.
    The optimization framework used to adjust this parameter is introduced in Appendix A of :cite:t:`paludEfficientSamplingNon2023`.
    """

    def __init__(
        self,
        list_lines: List[str],
        simu_name: str,
        D: int,
        D_no_kappa: int,
        K: int,
        log10_f_grid_size: int,
        N_samples_y: int,
        max_workers: int,
        sigma_a_raw: Union[np.ndarray, float],
        sigma_m: Union[np.ndarray, float],
        path_outputs: str,
        path_models: str,
        N_clusters_a_priori: Optional[int],
        small_size: int = 16,
        medium_size: int = 20,
        bigger_size: int = 24,
    ):
        r"""

        Parameters
        ----------
        list_lines : List[str]
            names of the observables for which the likelihood parameter needs to be adjusted
        simu_name : str
            simu_name of the process, to be used in the output folder simu_name
        D : int
            total number of physical parameters involved in the forward map
        D_no_kappa : int
            total number of physical parameters involved in the forward map except for the scaling parameter :math:`\kappa` (if it is part of the considered physical parameters)
        K : int
            the number of sampled theta values is :math:`K^D`
        log10_f_grid_size : int
            number of points in the grid on :math:`\log f_\ell(\theta)`
        N_samples_y : int
            number of samples for :math:`y_\ell`
        max_workers : int
            maximum number of workers that can be used for optimization or results extraction
        sigma_a_raw : Union[np.ndarray, float]
            standard deviation of the additive Gaussian noise
        sigma_m : Union[np.ndarray, float]
            standard deviation parameter of the multiplicative lognormal noise
        path_outputs : str
            path to the output folder (to be created), where the run results are to be saved
        path_models : str
            path to the folder containing the forward models
        N_clusters_a_priori: Optional[int]
            The number of different values of sigma_a_raw to consider for each line (and thus of optimization problem to solve per line), computed with a clustering algorithm on the `N` of values sigma_a_raw. Raises an error if `N_clusters_a_priori > N`.
        small_size : int, optional
            size for basic text, axes titles, xticks and yticks, by default 16
        medium_size : int, optional
            size of the axis labels, by default 20
        bigger_size : int, optional
            size of the figure title, by default 24
        """
        self.setup_plot_text_sizes(small_size, medium_size, bigger_size)
        self.create_empty_output_folders(simu_name, path_outputs)

        self.MODELS_PATH = path_models
        r"""str: path to the folder containing all the already defined and saved models (i.e., polynomials or neural networks)"""

        self.path_intermediate_result = (
            f"{self.path_output_sim}/best_params_intermediate.csv"
        )
        r"""str: path to the intermediate results (per cluster, saved in `output` folder)"""

        self.path_centroids = f"{self.path_output_sim}/sigma_a_centroids.csv"
        r"""str: path to the csv file containing the sigma_a centroid values for each cluster (saved in `output` folder)"""

        self.max_workers = max_workers
        r"""int: maximum number of workers that can be used for optimization or results extraction"""

        self.list_lines = list_lines
        r"""List[str]: names of the observables for which the likelihood parameter needs to be adjusted"""

        self.L = len(list_lines)
        r"""int: number of observables for which the likelihood parameter needs to be adjusted"""

        self.D = D
        r"""int: total number of physical parameters involved in the forward map"""

        self.D_no_kappa = D_no_kappa
        r"""int: total number of physical parameters involved in the forward map except for the scaling parameter :math:`\kappa` (if it is part of the considered physical parameters)"""

        self.K = K
        r"""int: the number of sampled theta values is ``K^D_sampling``"""

        self.N_samples_y = N_samples_y
        r"""int: number of samples for :math:`y_\ell`"""

        self.log10_f_grid_size = log10_f_grid_size  # nb points P(theta) grid
        r"""int: number of points in the grid on :math:`\log_{10} f_\ell(\theta)`"""

        self.N_samples_theta = 0
        r"""int: number of samples for :math:`\theta` used to build the histogram of :math:`\log_{10} f_\ell(\theta)`. To be defined in the daughter classes."""

        # check size of sigma_a_raw
        if isinstance(sigma_a_raw, np.ndarray):
            assert len(sigma_a_raw.shape) == 2, f"{sigma_a_raw.shape}"
            assert sigma_a_raw.shape[1] == self.L
            N = sigma_a_raw.shape[0]
        else:
            N = 1
            sigma_a_raw = sigma_a_raw * np.ones((N, self.L))

        self.N = N
        r"""int: number of pixels / components for which the optimization needs to be performed"""

        self.N_clusters = self.check_num_uniques(sigma_a_raw, N_clusters_a_priori)
        r"""Optional[int]: The number of different values of sigma_a to consider for each line (and thus of optimization problem to solve per line), computed with a clustering algorithm on the `N` of values sigma_a."""

        self.N_optim_per_line = self.N if self.N_clusters is None else self.N_clusters
        r"""int: number of optimization procedures to run per line"""

        print(
            f"run properties: N={self.N}, L={self.L}, N_clusters_a_priori={N_clusters_a_priori}, N_clusters={self.N_clusters}, N_optim={self.N_optim_per_line}\n"
        )

        # set sigma_a to sigma_a_raw if no clustering, else to centroids
        if self.N_optim_per_line == self.N:
            sigma_a = sigma_a_raw * 1
        else:
            # run the clustering algorithms on the set of standard deviations
            # for each line
            assert self.N_optim_per_line == self.N_clusters
            sigma_a = self.cluster_sigma_a_raw(sigma_a_raw)

        # check size of sigma_m
        assert isinstance(
            sigma_m, float
        ), "The current implementation only works for constant sigma_m over the map"
        if isinstance(sigma_m, np.ndarray):
            assert sigma_m.shape == (self.N_optim_per_line, self.L), f"{sigma_m.shape}"
        else:
            sigma_m = sigma_m * np.ones((self.N_optim_per_line, self.L))

        assert isinstance(sigma_a, np.ndarray)
        assert isinstance(sigma_m, np.ndarray)
        self.sigma_a = sigma_a
        r"""np.ndarray: standard deviations of the additive Gaussian noise"""

        self.sigma_m = sigma_m
        r"""np.ndarray: standard deviation parameter of the multiplicative lognormal noise"""

        return

    def check_num_uniques(
        self,
        sigma_a_raw: np.ndarray,
        N_clusters_a_priori: Optional[int],
    ) -> Optional[int]:
        r"""Sets the number of clusters to consider `N_clusters`, that is, the number of optimization procedure to run per line. There are 4 possible cases:

        * case 1: set N_clusters with the number of distinct values of sigma_a
        * case 2: the number of distinct values of sigma_a is lower than the number of clusters provided by the user. In this case, use the value that minimizes the number of optimization procedures to run.
        * case 3: use the number of clusters indicated by the user.
        * case 4: last case : run one optim per pixel, ie run self.N optimizations.

        Parameters
        ----------
        sigma_a_raw : np.ndarray of shape (N, L)
            set of standard deviations in the
        N_clusters_a_priori : Optional[int]
            number of clusters to consider indicated by the user

        Returns
        -------
        Optional[int]
            definitive number of clusters to consider
        """
        assert (N_clusters_a_priori is None) or (
            N_clusters_a_priori <= self.N
        ), "The number of clusters `N_clusters` should be either None or inferior to the total number of pixels."

        assert sigma_a_raw.shape == (
            self.N,
            self.L,
        ), f"sigma_a_raw has shape {sigma_a_raw.shape} but should have ({self.N}, {self.L})"

        # compute the number of distinct values for sigma_a in a single line
        max_distincts = 1
        for ell in range(self.L):
            n_uniques = np.unique(sigma_a_raw[:, ell]).size
            if max_distincts < n_uniques:
                max_distincts = n_uniques * 1

        # case 1 : set N_clusters with the number of distinct values of sigma_a
        if max_distincts < self.N and N_clusters_a_priori is None:
            N_clusters = max_distincts * 1
            return N_clusters

        # case 2 : the number of distinct values of sigma_a is lower than
        # the number of clusters provided by the user.
        # In this case, use the value that minimizes the number of
        # optimization procedures to run.
        if N_clusters_a_priori is not None and max_distincts < N_clusters_a_priori:
            N_clusters = max_distincts * 1
            return N_clusters

        # case 3 : use the number of clusters indicated by the user
        if N_clusters_a_priori is not None:
            return N_clusters_a_priori

        # last case : run one optim per pixel, ie run self.N optimizations
        return None

    @classmethod
    def parse_args(cls) -> Tuple[str, str, str, str]:
        """parses the inputs of the command-line, that should contain

        - the name of the input YAML file
        - path to the data folder
        - path to the models folder
        - path to the outputs folder to be created (by default '.')

        Returns
        -------
        str
            name of the input YAML file
        str
            path to the data folder
        str
            path to the models folder
        str
            path to the outputs folder to be created (by default '.')
        """
        if len(sys.argv) < 4:
            raise ValueError(
                "Please provide the following  arguments in your command: \n 1) the name of the input YAML file, \n 2) the path to the data folder, \n 3) the path to the models folder, \n 4) the path to the outputs folder to be created (by default '.')"
            )

        yaml_file = sys.argv[1]
        path_data = os.path.abspath(sys.argv[2])
        path_models = os.path.abspath(sys.argv[3])

        path_outputs = (
            os.path.abspath(sys.argv[4]) if len(sys.argv) == 5 else os.path.abspath(".")
        )
        path_outputs += "/outputs"

        print(f"input file name: {yaml_file}")
        print(f"path to data folder: {path_data}")
        print(f"path to models folder: {path_models}")
        print(f"path to outputs folder: {path_outputs}")

        return yaml_file, path_data, path_models, path_outputs

    def setup_plot_text_sizes(
        self,
        small_size: int = 16,
        medium_size: int = 20,
        bigger_size: int = 24,
    ) -> None:
        r"""defines text sizes on matplotlib plots

        Parameters
        ----------
        small_size : int, optional
            size for basic text, axes titles, xticks and yticks, by default 16
        medium_size : int, optional
            size of the axis labels, by default 20
        bigger_size : int, optional
            size of the figure title, by default 24
        """
        plt.rc("font", size=small_size)  # controls default text sizes
        plt.rc("axes", titlesize=small_size)  # fontsize of the axes title
        plt.rc("axes", labelsize=medium_size)  # fontsize of the x and y labels
        plt.rc("xtick", labelsize=small_size)  # fontsize of the tick labels
        plt.rc("ytick", labelsize=small_size)  # fontsize of the tick labels
        plt.rc("legend", fontsize=small_size)  # legend fontsize
        plt.rc("figure", titlesize=bigger_size)  # fontsize of the figure title
        return

    def create_empty_output_folders(self, simu_name: str, path_outputs: str) -> None:
        r"""creates the directories that receive the results of the likelihood parameter optimization

        Parameters
        ----------
        simu_name : str
            name of the simulation to be run
        path_yaml_file : str
            path of the folder containing the data and yaml files
        path_outputs: str
            folder where to write outputs
        """
        now = datetime.datetime.now()
        dt_str = now.strftime("%Y-%m-%d_%H")

        self.path_output_sim = os.path.abspath(
            f"{path_outputs}/approx_optim_{simu_name}_{dt_str}"
        )
        r"""str: path to the output root folder, e.g., ``./outputs/simu1``"""

        self.path_img = self.path_output_sim + "/img"
        r"""str: path to the output image folder, e.g., ``./outputs/simu1/img``"""
        self.path_img_hist = self.path_img + "/hist"
        r"""str: path to the histogram image folder, e.g., ``./outputs/simu1/img/hist``"""
        self.path_img_hist_final = self.path_img + "/hist_final"
        r"""str: path to the final histogram image folder, e.g., ``./outputs/simu1/img/hist_final``"""

        self.path_img_hist_sigma_a = self.path_img + "/hist_sigma_a"
        r"""str: path to the final histogram image folder, e.g., ``./outputs/simu1/img/hist_sigma_a``"""

        self.path_img_relation_sigma_a_approx = (
            self.path_img + "/relation_sigma_a_approx"
        )
        r"""str: path to the final histogram image folder, e.g., ``./outputs/simu1/img/relation_sigma_a_approx``"""

        self.path_img_final = self.path_img + "/final"
        r"""str: path to the final image folder, e.g., ``./outputs/simu1/img/final``"""

        self.path_logs = self.path_output_sim + "/logs"
        r"""str: path to the procedure logs, e.g., ``./outputs/logs``"""
        self.path_params = self.path_output_sim + "/optim_params"
        r"""str: path to the adjusted likelihood parameters, e.g., ``./outputs/optim_params``"""

        for folder_path in [
            path_outputs,
            self.path_output_sim,
            self.path_img,
            self.path_img_hist,
            self.path_img_hist_final,
            self.path_img_final,
            self.path_img_hist_sigma_a,
            self.path_img_relation_sigma_a_approx,
            self.path_logs,
            self.path_params,
        ]:
            if not os.path.isdir(folder_path):
                os.mkdir(folder_path)

        return

    def sample_theta(
        self,
        lower_bounds: np.ndarray,
        upper_bounds: np.ndarray,
    ) -> np.ndarray:
        r"""sample :math:`\theta` from Stratified MC in cube

        Parameters
        ----------
        K : int
            total number of samples per axis
        lower_bounds : np.ndarray of shape (D,)
            lower bounds of cube on :math:`\theta`
        upper_bounds : np.ndarray of shape (D,)
            upper bounds of cube on :math:`\theta`

        Returns
        -------
        np.ndarray of shape (N_samples, D)
            :math:`\theta` samples
        """
        print("starting generation of theta values with Stratified MC")
        rng = np.random.default_rng(42)
        D_sampling: int = upper_bounds.size

        delta_upper_lower = np.array(
            [upper_bounds[d] - lower_bounds[d] for d in range(D_sampling)]
        )
        list_lower_bounds = [
            np.arange(self.K) / self.K * delta_upper_lower[d] + lower_bounds[d]
            for d in range(D_sampling)
        ]
        list_lower_bounds = np.meshgrid(*list_lower_bounds)
        list_lower_bounds = [
            lower_bounds_Theta_i.flatten() for lower_bounds_Theta_i in list_lower_bounds
        ]
        list_samples = []
        for d, lower_bounds_Theta_i in enumerate(list_lower_bounds):
            Vd = rng.uniform(
                low=lower_bounds_Theta_i,
                high=lower_bounds_Theta_i + delta_upper_lower[d] / self.K,
            )
            list_samples.append(Vd)
        x = np.vstack(list_samples).T
        print("generation of theta values with Stratified MC done")
        return x

    def plot_hist_log10_f_Theta(
        self,
        log10_f_Theta: np.ndarray,
        log10_f_Theta_low: float,
        log10_f_Theta_high: float,
        list_log10_f_grid: np.ndarray,
        pdf_kde_log10_f_Theta: np.ndarray,
        ell: int,
    ) -> None:
        r"""plots histogram of :math:`log_{10}(f_\ell(\theta))`

        Parameters
        ----------
        log10_f_Theta : np.ndarray of shape (-1, 1)
            array of values of :math:`\log_{10} f_\ell (\theta)` for considered line
        log10_f_Theta_low: float
            lower bound for :math:`\log_{10} f_\ell (\theta)` for the considered line
        log10_f_Theta_high: float
            upper bound for :math:`\log_{10} f_\ell (\theta)` for the considered line
        list_log10_f_grid: np.ndarray
            grid values of :math:`\log_{10} f_\ell (\theta)` for the considered line
        pdf_kde_log10_f_Theta: np.ndarray
            pdf of :math:`\log_{10} f_\ell (\theta)` evaluated with a kernel density estimator
        ell : int
            index of the line
        """
        plt.figure(figsize=(8, 6))
        plt.title(f"line {ell} : {self.list_lines[ell]}")
        plt.hist(
            log10_f_Theta,
            bins=self.log10_f_grid_size,
            range=(log10_f_Theta_low, log10_f_Theta_high),
            density=True,
            label="samples",
        )
        plt.plot(list_log10_f_grid, pdf_kde_log10_f_Theta, "k--", label="KDE")
        plt.xlabel(r"$\log_{10} f_\ell (\theta)$")
        plt.grid()
        plt.legend()
        plt.tight_layout()
        plt.savefig(f"{self.path_img_hist}/hist_log_f_Theta_{self.list_lines[ell]}.PNG")
        plt.close()

        return

    def plot_hist_log10_f_Theta_with_optim_results(
        self,
        log10_f_Theta: np.ndarray,
        log10_f_Theta_low: float,
        log10_f_Theta_high: float,
        list_log10_f_grid: np.ndarray,
        pdf_kde_log10_f_Theta: np.ndarray,
        n: int,
        ell: int,
        best_point: np.ndarray,
    ) -> None:
        r"""plots histogram of :math:`\log_{10} f_\ell (\theta)`

        Parameters
        ----------
        log10_f_Theta : np.ndarray of shape (-1, 1)
            array of values of :math:`\log_{10} f_\ell (\theta)` for considered line
        log10_f_Theta_low: float
            lower bound for :math:`\log_{10} f_\ell (\theta)` for the considered line
        log10_f_Theta_high: float
            upper bound for :math:`\log_{10} f_\ell (\theta)` for the considered line
        list_log10_f_grid: np.ndarray
            grid values of :math:`\log_{10} f_\ell (\theta)` for the considered line
        pdf_kde_log10_f_Theta: np.ndarray
            pdf of :math:`\log_{10} f_\ell (\theta)` evaluated with a kernel density estimator
        n : int
            pixel / component index
        ell : int
            index of the line
        best_point : np.ndarray
            position for the best point, to be displayed
        """
        lambda_ = 1 - utils.compute_lambda(
            a0=best_point[0],
            a1=best_point[1],
            f_Theta_true=np.exp(list_log10_f_grid * np.log(10.0)),
        )

        var_eps_m = np.exp(self.sigma_m**2) * (np.exp(self.sigma_m**2) - 1)
        log10_f0_n_ell = (
            0.5
            * (2 * np.log(self.sigma_a[n, ell]) - np.log(var_eps_m[n, ell]))
            / np.log(10)
        )

        plt.figure(figsize=(8, 6))
        plt.title(f"line {ell} : {self.list_lines[ell]}")
        vals, _, _ = plt.hist(
            log10_f_Theta,
            bins=self.log10_f_grid_size,
            range=(log10_f_Theta_low, log10_f_Theta_high),
            density=True,
            label="samples",
        )
        plt.plot(list_log10_f_grid, pdf_kde_log10_f_Theta, "k--", label="KDE")
        plt.plot(
            list_log10_f_grid,
            lambda_ * np.max(vals),
            "r-",
            label=r"multi. weight $\lambda$",
        )
        plt.axvline(
            np.log10(self.sigma_a[n, ell]), c="orange", ls="-", label=r"$\sigma_a$"
        )
        plt.axvline(
            log10_f0_n_ell,
            c="orange",
            ls="--",
            label=r"$\sigma_a^2 = f_\ell(\theta)^2$ Var$( \epsilon^{(m)}_{\ell} )$",
        )

        plt.xlabel(r"$\log_{10} f_\ell (\theta)$")
        plt.grid()
        plt.legend()
        plt.tight_layout()
        plt.savefig(
            f"{self.path_img_hist_final}/hist_final_log_f_Theta_n{n}_{self.list_lines[ell]}.PNG"
        )
        plt.close()

        return

    def save_setup_to_json(
        self, n: int, ell: int, pbounds: Dict[str, np.ndarray]
    ) -> None:
        """save optimization context and parameters

        Parameters
        ----------
        n : int
            pixel / component index
        ell : int
            observable index
        pbounds : dict[str, np.ndarray]
            contains the bounds on the parameters to be adjusted
        """
        optim_params = {
            "n": n,
            "ell": ell,
            "line": self.list_lines[ell],
            "sigma_a": self.sigma_a[n, ell],
            "sigma_m": self.sigma_m[n, ell],
            "K": self.K,
            "N_samples_theta": self.N_samples_theta,
            "N_samples_y": self.N_samples_y,
            "log10_f_grid_size": self.log10_f_grid_size,
            "a0": {"low": pbounds["a0"][0], "high": pbounds["a0"][1]},
            "a1": {"low": pbounds["a1"][0], "high": pbounds["a1"][1]},
        }
        with open(
            f"{self.path_params}/optim_params_n{n}_{self.list_lines[ell]}.json",
            "w",
            encoding="utf-8",
        ) as f:
            json.dump(
                optim_params,
                f,
                ensure_ascii=False,
                indent=4,
            )
        return

    def rewrite_logs_correct_json_format(self) -> None:
        r"""rewrites the log files with correct json format"""
        for ell in range(self.L):
            line = self.list_lines[ell] * 1

            for n in range(self.N_optim_per_line):
                with open(f"{self.path_logs}/logs_n{n}_{line}.json", "r") as f:
                    text = f.read()

                text = "[" + text.replace("}}", "}},")[:-2] + "]"
                with open(
                    f"{self.path_logs}/logs_correct_format_n{n}_{line}.json",
                    "w",
                ) as f:
                    f.write(text)

        return

    def extract_optimal_params(self) -> pd.DataFrame:
        """extracts the adjusted likelihood parameters from the log files and gather them in a DataFrame

        Returns
        -------
        pd.DataFrame
            DataFrame with the set of evaluated optimal parameters for each component n and line ell
        """
        self.rewrite_logs_correct_json_format()

        list_best = []
        for n in range(self.N_optim_per_line):
            for ell in range(self.L):
                line = self.list_lines[ell] * 1
                # import data
                with open(
                    f"{self.path_logs}/logs_correct_format_n{n}_{line}.json",
                    "r",
                ) as f:
                    data = json.load(f)

                # Flatten data
                df = pd.json_normalize(data, max_level=1)
                df = df.rename(columns={"params.a0": "a0", "params.a1": "a1"})

                # get best
                df_sorted = df.sort_values("target", ascending=False)
                idx_best = df_sorted.index[0]
                dict_best = {
                    "n": n,
                    "ell": ell,
                    "line": self.list_lines[ell],
                    "sigma_a": self.sigma_a[n, ell],
                    "sigma_m": self.sigma_m[n, ell],
                    "a0_best": df.at[idx_best, "a0"],
                    "a1_best": df.at[idx_best, "a1"],
                    "target_best": df.at[idx_best, "target"],
                }
                list_best.append(dict_best)

        # * save best points
        df_best = pd.DataFrame(list_best)
        df_best.to_csv(self.path_intermediate_result, index=False)
        return df_best

    def setup_params_bounds(
        self,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, float, float]:
        r"""sets the bounds on the parameters to be adjusted, defined here as transition interval center for ``a0`` position and width for ``a1``

        Returns
        -------
        log10_f0 : np.ndarray of shape (N, L)
            values of :math:`f_\ell(\theta)` at which additive and multiplicative noise variances are equal
        bounds_a0_low : np.ndarray of shape (N, L)
            lower bounds on the center of the transition interval (defined as deltas around the ``log10_f0``)
        bounds_a0_high : np.ndarray of shape (N, L)
            upper bounds on the center of the transition interval (defined as deltas around the ``log10_f0``)
        bounds_a1_low : float
            lower value for the transition interval size
        bounds_a1_high : float
            upper value for the transition interval size
        """
        print("starting definition of parameters bounds")
        # log10_f0 = log10(ratio std of noises)
        # ie value of f at which noise variances are equal
        var_eps_m: np.ndarray = np.exp(self.sigma_m**2) * (
            np.exp(self.sigma_m**2) - 1
        )
        log10_f0: np.ndarray = (
            0.5 * (2 * np.log(self.sigma_a) - np.log(var_eps_m)) / np.log(10)
        )

        bounds_a0_low: np.ndarray = log10_f0 - 2  # (N, L)
        bounds_a0_high: np.ndarray = log10_f0 + 8  # (N, L)

        bounds_a1_low = 0.01
        bounds_a1_high = 2.0
        print("definition of parameters bounds done")

        return (
            log10_f0,
            bounds_a0_low,
            bounds_a0_high,
            bounds_a1_low,
            bounds_a1_high,
        )

    def cluster_sigma_a_raw(self, sigma_a_raw: np.ndarray) -> np.ndarray:
        r"""runs `self.L` k-means clustering algorithms (one per line) on the  sets of standard deviation of the additive noise `sigma_a`. The number of clusters is defined with `self.N_clusters`. The obtained dataframe is saved as a `csv` file.

        Parameters
        ----------
        sigma_a_raw : np.ndarray of shape (N, L)
            array of all the sigma_a values associated with the observations.

        Returns
        -------
        np.ndarray of shape (N_clusters, L)
            reduced sigma_a array, with `N_clusters` lines instead of `N` (with `N_clusters` potentially much smaller than `N`)
        """
        assert self.N_clusters is not None

        df_clusters = pd.DataFrame()
        df_clusters["Y"] = np.arange(self.N_clusters)
        df_clusters["X"] = 0

        for ell, line in enumerate(self.list_lines):
            log10_sigma_a_ell_nonnan = np.log10(sigma_a_raw[:, ell])
            log10_sigma_a_ell_nonnan = log10_sigma_a_ell_nonnan[
                ~np.isnan(log10_sigma_a_ell_nonnan)
            ]
            log10_sigma_a_ell_nonnan = log10_sigma_a_ell_nonnan[
                log10_sigma_a_ell_nonnan < 3.0
            ]

            max_min_diff = (
                log10_sigma_a_ell_nonnan.max() - log10_sigma_a_ell_nonnan.min()
            )

            km = KMeans(self.N_clusters).fit(log10_sigma_a_ell_nonnan.reshape((-1, 1)))
            log10_centroids = km.cluster_centers_.flatten()
            log10_centroids.sort()
            df_clusters[line] = np.exp(log10_centroids * np.log(10))

            self.plot_clusters_sigma_a(line, log10_sigma_a_ell_nonnan, log10_centroids)

            assert (
                max_min_diff < 15.0
            ), f"The difference between the min and max is over 15 orders of magnitude for line {line}. This should not happen."

        df_clusters = df_clusters.set_index(["X", "Y"])
        df_clusters.to_csv(self.path_centroids)

        return df_clusters.values

    def plot_clusters_sigma_a(
        self,
        line: str,
        log10_sigma_a_ell_nonnan: np.ndarray,
        log10_centroids: np.ndarray,
    ) -> None:
        r"""plots one one-dimension histogram on the log10 of the standard deviation on the additive noise in the observation, and the computed centroids.

        Parameters
        ----------
        line : str
            name of the line
        log10_sigma_a_ell_nonnan : np.ndarray
            non-nan values of sigma_a for the considered line
        log10_centroids : np.ndarray of shape (N_clusters,)
            values of the sigma_a centroids for the considered line
        """
        assert self.N_clusters is not None
        assert (
            isinstance(log10_sigma_a_ell_nonnan, np.ndarray)
            and len(log10_sigma_a_ell_nonnan.shape) == 1
        )
        assert (
            isinstance(log10_centroids, np.ndarray)
            and log10_centroids.size == self.N_clusters
        )

        plt.title(f"line {line}: sigma_a and centroids")
        plt.hist(log10_sigma_a_ell_nonnan, bins=50)

        for j in range(self.N_clusters):
            if j == 0:
                plt.axvline(log10_centroids[j], ls="--", c="k", label="centroids")
            else:
                plt.axvline(log10_centroids[j], ls="--", c="k")

        plt.ylabel("counts in sigma_a")
        plt.xlabel(
            r"$\log_{10} \sigma_{a,n\ell}$ for all pixels $n$ and one line $\ell$"
        )

        plt.legend()
        plt.tight_layout()
        plt.savefig(f"{self.path_img_hist_sigma_a}/sigma_a_and_centroids_{line}.PNG")
        plt.close()

    def save_results_in_data_folder(self, path_data: str, filename_err: str) -> None:
        r"""

        Parameters
        ----------
        path_data : str
            path to the data folder
        filename_err : str
            name of the file containing the sigma_a values
        """
        filename_final_result = f"{path_data}/best_params_approx.csv"

        if self.N_optim_per_line == self.N_clusters:
            # step 1 : aller chercher le best_params.csv dans l'output folder
            df_results = pd.read_csv(
                self.path_intermediate_result, index_col=["n", "line"]
            )

            # step 2 : create the large best_params.csv file
            # in case of nan, set a1_best = 3.0, a0_best=0.01
            df_err = pd.read_pickle(f"{path_data}/{filename_err}")
            df_err["n"] = np.arange(len(df_err))
            df_err = df_err.reset_index().set_index("n")

            df_centroids = pd.read_csv(self.path_centroids)

            cols_results = ["a0_best", "a1_best", "target_best"]

            df_best_params_full = pd.DataFrame(
                columns=cols_results,
                index=pd.MultiIndex.from_product(
                    [np.arange(self.N), self.list_lines], names=["n", "line"]
                ),
            )

            for line in self.list_lines:
                centroids = np.log10(df_centroids.loc[:, line].values)

                for n in range(self.N):
                    # get sigma_{a,n\ell}
                    log10_sigma_a = np.log10(df_err.at[n, line])

                    if np.isnan(log10_sigma_a):
                        df_best_params_full.loc[(n, line), cols_results] = [
                            3.0,
                            0.01,
                            np.nan,
                        ]
                    else:
                        idx_min = np.argmin(np.abs(log10_sigma_a - centroids))

                        df_best_params_full.loc[(n, line), cols_results] = (
                            df_results.loc[(idx_min, line), cols_results] * 1
                        )

            # step 3 : save it in the same folder as the input_params.yaml file
            df_best_params_full.to_csv(filename_final_result)

        else:
            # copy and paste the re
            assert self.N_optim_per_line == self.N
            shutil.copyfile(self.path_intermediate_result, filename_final_result)
        return

    def plot_params_with_sigma_a(self, df_best: pd.DataFrame) -> None:
        for ell, line in enumerate(self.list_lines):
            a0 = df_best.loc[df_best["line"] == line, "a0_best"]
            a1 = df_best.loc[df_best["line"] == line, "a1_best"]

            fig, ax = plt.subplots(
                1, 2, sharex=True, constrained_layout=True, figsize=(6, 3)
            )

            fig.suptitle(f"line {line}")
            ax[0].plot(np.log10(self.sigma_a[:, ell]), a0, "+")
            ax[0].set_ylabel("a0 best")
            ax[0].set_xlabel(r"$\log_{10} \sigma_a$")
            ax[0].grid()

            ax[1].plot(np.log10(self.sigma_a[:, ell]), a1, "+")
            ax[1].set_ylabel("a1 best")
            ax[1].set_xlabel(r"$\log_{10} \sigma_a$")
            ax[1].grid()

            plt.savefig(f"{self.path_img_relation_sigma_a_approx}/{line}.PNG")
            plt.close()
        return



"""Implementation of an approximation of the likelihood function of a mixture of Gaussian and multiplicative noises with censorship with a lower limit.
"""
from typing import List, Optional, Union

import numba
import numpy as np
import pandas as pd
from scipy.special import log_ndtr

from beetroots.modelling.likelihoods import utils
from beetroots.modelling.likelihoods.abstract_likelihood import Likelihood


@numba.njit()
def cost_au(
    y: np.ndarray, f_Theta: np.ndarray, m_a: np.ndarray, s_a: np.ndarray
) -> np.ndarray:
    nll_au = 0.5 * ((y - f_Theta - m_a) / s_a) ** 2 + np.log(s_a)
    return nll_au


@numba.njit()
def cost_mu(
    log_y: np.ndarray, log_f_Theta: np.ndarray, m_m: np.ndarray, s_m: np.ndarray
) -> np.ndarray:
    nll_mu = 0.5 * ((log_y - log_f_Theta - m_m) / s_m) ** 2 + np.log(s_m)
    # nll_mu = np.where(np.isnan(log_y) | np.isinf(log_y), 10.0 ** 8, nll_mu)

    # shape = nll_mu.shape
    # nll_mu = nll_mu.ravel()
    # nll_mu[np.isnan(nll_mu)] = 0
    # nll_mu = nll_mu.reshape(shape)
    return nll_mu


@numba.njit()
def gradient_cost_au(
    y: np.ndarray,
    grad_f_Theta: np.ndarray,
    f_Theta: np.ndarray,
    grad_m_a: np.ndarray,
    m_a: np.ndarray,
    grad_s_a2: np.ndarray,
    s_a2: np.ndarray,
) -> np.ndarray:
    u_1 = (f_Theta - m_a - y) / s_a2**2
    assert u_1.shape == f_Theta.shape  # (N, L)

    u_2 = (
        2 * np.expand_dims(s_a2, axis=1) * (grad_f_Theta + grad_m_a)
        - np.expand_dims(f_Theta - m_a - y, axis=1) * grad_s_a2
    )
    assert u_2.shape == grad_f_Theta.shape  # (N, D, L)

    g_au = 0.5 * grad_s_a2 / np.expand_dims(s_a2, axis=1)
    g_au += 0.5 * np.expand_dims(u_1, axis=1) * u_2

    assert g_au.shape == grad_f_Theta.shape
    return g_au


@numba.njit()
def gradient_cost_mu(
    log_y: np.ndarray,
    grad_log_f_Theta: np.ndarray,
    log_f_Theta: np.ndarray,
    grad_m_m: np.ndarray,
    m_m: np.ndarray,
    grad_s_m2: np.ndarray,
    s_m2: np.ndarray,
) -> np.ndarray:

    u_1 = (log_f_Theta - m_m - log_y) / s_m2**2
    assert u_1.shape == log_f_Theta.shape  # (N, L)

    u_2 = (
        2 * np.expand_dims(s_m2, axis=1) * (grad_log_f_Theta + grad_m_m)
        - np.expand_dims(log_f_Theta - m_m - log_y, axis=1) * grad_s_m2
    )
    assert u_2.shape == grad_log_f_Theta.shape  # (N, D, L)

    g_mu = 0.5 * grad_s_m2 / np.expand_dims(s_m2, axis=1)
    g_mu += np.where(
        np.expand_dims(np.isnan(log_y), axis=1) + np.zeros_like(grad_log_f_Theta),
        # fake output
        np.zeros_like(grad_log_f_Theta),
        # true grad
        0.5 * np.expand_dims(u_1, axis=1) * u_2,
    )

    assert g_mu.shape == grad_log_f_Theta.shape
    return g_mu


class MixingModelsLikelihood(Likelihood):
    r"""Class implementing a Gaussian likelihood model with lower censorship. This likelihood function is introduced in Section II.C of :cite:t:`paludEfficientSamplingNon2023`.

    Note
    ----
    This likelihood is a parametric approximation of the true likelihood model.
    The associated parameter, denoted :math:`a_\ell` in the article (as there is one such parameter per observable :math:`\ell `), should be adjusted before any inversion.
    To adjust this parameter, see the ``beetroots.approx_optim`` subpackage.
    """

    __slots__ = (
        "N",
        "L",
        "D",
        "y",
        "forward_map",
        "log_y",
        "sigma_a",
        "sigma_m",
        "omega",
        "log_fm1",
        "log_fp1",
        "P_lambda",
        "grad_P_lambda",
        "hess_diag_P_lambda",
    )

    def __init__(
        self,
        forward_map,
        D: int,
        L: int,
        N: int,
        y: np.ndarray,
        sigma_a: Union[float, np.ndarray],
        sigma_m: Union[float, np.ndarray],
        omega: Union[float, np.ndarray],
        path_transition_params: str,
        list_lines_fit: List[str],
    ) -> None:
        """Constructor of the GaussianLikelihood object.

        Parameters
        ----------
        forward_map : ForwardMap instance
            forward map
        D : int
            number of disinct physical parameters in input space.
        L : int
            number of distinct observed physical parameters.
        N : int
            number of pixels in each physical dimension
        y : np.ndarray of shape (N, L)
            mean of the gaussian distribution
        sigma_a : float or np.ndarray of shape (N, L)
            standard deviation of the Gaussian distribution
        sigma_m : float or np.ndarray of shape (N, L)
            scale parameter of the lognormal distribution
        omega : float or np.ndarray of shape (N, L)
            censorship threshold

        Raises
        ------
        ValueError
            y must have the shape (N, L)
        """
        super().__init__(forward_map, D, L, N, y)

        # ! trigger an error is the mean y contains less than N elements
        if not (y.shape == (N, L)):
            raise ValueError(
                "y must have the shape (N, L) = ({}, {}) elements".format(
                    self.N, self.L
                )
            )
        if isinstance(sigma_a, (float, int)):
            self.sigma_a = sigma_a * np.ones((N, L))
        else:
            assert sigma_a.shape == (N, L)
            self.sigma_a = sigma_a

        if isinstance(sigma_m, (float, int)):
            self.sigma_m = sigma_m * np.ones((N, L))
        else:
            assert sigma_m.shape == (N, L)
            self.sigma_m = sigma_m

        if isinstance(omega, (float, int)):
            self.omega = omega * np.ones((N, L))
        else:
            assert omega.shape == (N, L)
            self.omega = omega

        self.log_y = np.log(y)

        self.set_likelihood_approx_parameters(path_transition_params, list_lines_fit)

        # set polynom for transition
        self.P_lambda = np.poly1d(np.array([-6.0, 15.0, -10.0, 0.0, 0.0, 1.0]))
        self.grad_P_lambda = self.P_lambda.deriv(m=1)
        self.hess_diag_P_lambda = self.P_lambda.deriv(m=2)

    def set_likelihood_approx_parameters(
        self, path_transition_params: str, list_lines_fit: List[str]
    ) -> None:
        r"""Sets the likelihood approximation parameters by reading them from the csv file indicated in the `mixing_model_params_filename` entry of the input params yaml file. This `mixing_model_params_filename` file is obtained with the Bayesian Optimization procedure implemented in the `approx_optim` subpackage (see documentation for more details on how to obtain this file).

        The parameters are called `log_fm1` and `log_fp1`. They correspond to the lower and upper limits of the interval in which the two approximations (additive and multiplicative) are combined (below the lower limit, only the additive model is used, and above the upper limit, only the multiplicative model is used).

        These two parameters are computed for each pixel `n` and line `ell` from the two parameters that are stored in the `.csv` file, which are the center and the radius of the interval.

        Parameters
        ----------
        path_transition_params : str
            path to the file that contains the approximation parameters
        list_lines_fit : List[str]
            list of lines used during the fit
        """
        df_transition = pd.read_csv(path_transition_params)
        num_distinct_n = np.unique(df_transition["n"]).size

        # check that the lines in list_lines_fit are in the
        # df_transition dataframe
        list_unique_lines = np.unique(df_transition["line"].values)
        for line in list_lines_fit:
            assert (
                line in list_unique_lines
            ), f"the line {line} is not in the file containing the likelihood approximation parameters"

        # in case there is only one set of approx parameters per line (ie, to be used for each pixel)
        if num_distinct_n == 1:
            df_transition = df_transition.set_index("line")

            transition_center = (
                df_transition.loc[list_lines_fit, "a0_best"].values[None, :]
                * np.ones((self.N, self.L))
                * np.log(10)
            )
            transition_radius = (
                df_transition.loc[list_lines_fit, "a1_best"].values[None, :]
                * np.ones((self.N, self.L))
                * np.log(10)
            )
            print(
                "Using the same set of approximation parameters for all pixels. Careful: this can lead to estimation errors."
            )

        # in case there are as many sets of approx parameters per line as pixels
        else:
            assert (
                num_distinct_n == self.N
            ), f"the transition file {path_transition_params} should have {self.N} or 1 distinct values of n, and has {num_distinct_n}"

            df_transition = df_transition.set_index(["n", "line"])
            index = pd.MultiIndex.from_product(
                [list(range(self.N)), list_lines_fit],
                names=["n", "line"],
            )

            transition_center = df_transition.loc[index, "a0_best"].values.reshape(
                (self.N, self.L)
            ) * np.log(10)
            transition_radius = df_transition.loc[index, "a1_best"].values.reshape(
                (self.N, self.L)
            ) * np.log(10)
            print(
                "Using sets of approximation parameters defined for each pixel and line."
            )

        self.log_fm1 = transition_center - transition_radius  # (N, L)
        self.log_fp1 = transition_center + transition_radius  # (N, L)
        return

    def sample_observation_model(
        self,
        forward_map_evals: dict,
        rng: np.random.Generator = np.random.default_rng(),
    ) -> np.ndarray:
        eps_a = rng.normal(loc=0.0, scale=self.sigma_a)
        eps_m = rng.lognormal(
            mean=-(self.sigma_m**2) / 2,
            sigma=self.sigma_m,
        )
        f_Theta = forward_map_evals["f_Theta"] * 1
        y_rep = np.maximum(self.omega, eps_m * f_Theta + eps_a)
        return y_rep

    def model_mixing_param(
        self,
        forward_map_evals: dict,
        idx: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""computes the weight of the additive model :math:`\lambda_{n, \ell}` (line-wise and pixel-wise). In this model, :math:`\lambda_{n, \ell}` is a functino of the observation :math:`y_{n, \ell}`, and therefore constant during the sampling

        .. math::
            \lambda_{n, \ell} = \frac{\sigma_a^{2b}}{\sigma_a^{2b} + (a \sigma_m y_{n, \ell})^{2b}}

        with :math:`a` a transition location parameter and :math:`b` a transition speed parameter
        """
        if idx is None:
            N_pix = self.N * 1
            log_fm1 = self.log_fm1 * 1
            log_fp1 = self.log_fp1 * 1
            # sigma_a = self.sigma_a * 1
            # sigma_m = self.sigma_m * 1
        else:
            n_pix = idx.size
            k_mtm = forward_map_evals["f_Theta"].shape[0] // n_pix
            N_pix = forward_map_evals["f_Theta"].shape[0]

            log_fm1 = np.zeros((n_pix, k_mtm, self.L))
            log_fp1 = np.zeros((n_pix, k_mtm, self.L))

            for i_pix in range(n_pix):
                log_fm1[i_pix, :, :] = self.log_fm1[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                log_fp1[i_pix, :, :] = self.log_fp1[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )

            log_fm1 = log_fm1.reshape((N_pix, self.L))
            log_fp1 = log_fp1.reshape((N_pix, self.L))

            # N_pix = forward_map_evals["f_Theta"].shape[0]

            # log_fm1 = np.zeros((N_pix, self.L))
            # log_fp1 = np.zeros((N_pix, self.L))
            # for i in range(idx.size):
            #     log_fm1[i, :] = self.log_fm1[idx[i], :] * 1
            #     log_fp1[i, :] = self.log_fp1[idx[i], :] * 1

            # log_fp1 = self.log_fp1[idx, :][None, :] * np.ones((N_pix, self.L))

            # sigma_a = self.sigma_a[idx, :][None, :] * np.ones((N_pix, self.L))
            # sigma_m = self.sigma_m[idx, :][None, :] * np.ones((N_pix, self.L))

        # lambda_ = 1 / (
        #     1
        #     + np.exp(
        #         (2 * self.transition_speed)
        #         * np.log(
        #             (self.transition_loc * sigma_m * forward_map_evals["f_Theta"]) / sigma_a
        #         )
        #     )
        # )
        lambda_ = np.where(
            forward_map_evals["log_f_Theta"] <= log_fm1,
            1,
            np.where(
                forward_map_evals["log_f_Theta"] >= log_fp1,
                0,
                self.P_lambda(
                    (forward_map_evals["log_f_Theta"] - log_fm1) / (log_fp1 - log_fm1)
                ),
            ),
        )
        return lambda_  # (N, L)

    def grad_model_mixing_param(self, forward_map_evals: dict) -> np.ndarray:
        r"""[summary]

        Parameters
        ----------
        lambda_ : np.ndarray of shape (N, L)
            [description]
        f_Theta : np.ndarray of shape (N, L)
            [description]
        grad_f_Theta : np.ndarray of shape (N, D, L)
            [description]

        Returns
        -------
        np.ndarray of shape (N, D, L)
            [description]
        """
        # grad_ = (
        #     (-2 * self.transition_speed)
        #     * (
        #         (
        #             (self.transition_loc * self.sigma_m / self.sigma_a)
        #             ** (2 * self.transition_speed)
        #         )
        #         * (
        #             forward_map_evals["f_Theta"] ** (2 * self.transition_speed - 1)
        #             * lambda_ ** 2
        #         )
        #     )[:, None, :]
        #     * forward_map_evals["grad_f_Theta"]
        # )
        u = (forward_map_evals["log_f_Theta"] - self.log_fm1) / (
            self.log_fp1 - self.log_fm1
        )  # (N, L)
        grad_u = (
            forward_map_evals["grad_log_f_Theta"]
            / (self.log_fp1 - self.log_fm1)[:, None, :]
        )  # (N, D, L)
        # assert u.shape == (self.N, self.L)
        # assert grad_u.shape == (self.N, self.D, self.L), grad_u.shape

        grad_ = np.where(
            (
                (forward_map_evals["log_f_Theta"] <= self.log_fm1)
                | (forward_map_evals["log_f_Theta"] >= self.log_fp1)
            )[:, None, :],
            np.zeros((self.N, self.D, self.L)),
            grad_u * self.grad_P_lambda(u)[:, None, :],
        )
        assert grad_.shape == (self.N, self.D, self.L), grad_.shape

        return grad_  # (N, D, L)

    def hess_diag_model_mixing_param(self, forward_map_evals: dict) -> np.ndarray:
        r"""[summary]

        Parameters
        ----------
        lambda_ : np.ndarray of shape (N, L)
            [description]
        grad_lambda_ : np.ndarray of shape (N, D, L)
            [description]
        f_Theta : np.ndarray of shape (N, L)
            [description]
        grad_f_Theta : np.ndarray of shape (N, D, L)
            [description]
        hess_diag_f_Theta : np.ndarray of shape (N, D, L)
            [description]

        Returns
        -------
        np.ndarray of shape (N, D, L)
            [description]
        """
        # hess_diag = (
        #     (-2 * self.transition_speed)
        #     * (self.transition_loc * self.sigma_m / self.sigma_a)
        #     ** (2 * self.transition_speed)
        # )[:, None, :] * (
        #     (
        #         forward_map_evals["f_Theta"] ** (2 * self.transition_speed - 1)
        #         * lambda_ ** 2
        #     )[:, None, :]
        #     * forward_map_evals["hess_diag_f_Theta"]
        #     + (2 * self.transition_speed - 1)
        #     * (
        #         forward_map_evals["grad_f_Theta"] ** 2
        #         * (
        #             lambda_ ** 2
        #             * forward_map_evals["f_Theta"] ** (2 * self.transition_speed - 2)
        #         )[:, None, :]
        #     )
        #     + 2
        #     * (forward_map_evals["f_Theta"] ** (2 * self.transition_speed - 1) * lambda_)[
        #         :, None, :
        #     ]
        #     * forward_map_evals["grad_f_Theta"]
        #     * grad_lambda_
        # )

        u = (forward_map_evals["log_f_Theta"] - self.log_fm1) / (
            self.log_fp1 - self.log_fm1
        )  # (N, L)
        grad_u = (
            forward_map_evals["grad_log_f_Theta"]
            / (self.log_fp1 - self.log_fm1)[:, None, :]
        )  # (N, D, L)
        hess_diag_u = (
            forward_map_evals["hess_diag_log_f_Theta"]
            / (self.log_fp1 - self.log_fm1)[:, None, :]
        )  # (N, D, L)

        hess_diag = np.where(
            (
                (forward_map_evals["log_f_Theta"] <= self.log_fm1)
                | (forward_map_evals["log_f_Theta"] >= self.log_fp1)
            )[:, None, :],
            np.zeros((self.N, self.D, self.L)),
            (
                grad_u**2 * self.hess_diag_P_lambda(u)[:, None, :]
                + hess_diag_u * self.grad_P_lambda(u)[:, None, :]
            ),
        )
        assert hess_diag.shape == (self.N, self.D, self.L), hess_diag.shape
        return hess_diag  # (N, D, L)

    def _compute_bias_and_std(
        self,
        forward_map_evals: dict,
        idx: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""computes the biases and std of additive and multiplicative approximations

        Parameters
        ----------
        f_Theta : numpy.array of shape (N_pix, L)
            image of the current iterate through the forward model
        sigma_a : np

        Returns
        -------
        current_bias_std : np.ndarray of shape (4, N_pix, L)
            array with (m_a, s_a, m_m, s_m)
        """
        if idx is None:
            N_pix = self.N * 1
            sigma_a = self.sigma_a * 1
            sigma_m = self.sigma_m * 1

        else:
            N_pix = forward_map_evals["f_Theta"].shape[0]

            sigma_a = np.zeros((N_pix, self.L))
            sigma_m = np.zeros((N_pix, self.L))
            for i in range(idx.size):
                sigma_a[i, :] = self.sigma_a[idx[i], :] * 1
                sigma_m[i, :] = self.sigma_m[idx[i], :] * 1

        log_combination = (
            np.log(sigma_a) - forward_map_evals["log_f_Theta"] - (sigma_m**2) / 2
        )

        assert sigma_a.min() > 0, sigma_a.min()
        assert sigma_m.min() > 0, sigma_m.min()
        assert np.sum(np.isnan(forward_map_evals["log_f_Theta"])) == 0, np.sum(
            np.isnan(forward_map_evals["log_f_Theta"])
        )
        assert np.sum(np.isnan(log_combination)) == 0, np.sum(np.isnan(log_combination))

        # * computation of bias and variances
        m_a = (np.exp(sigma_m**2 / 2) - 1) * forward_map_evals["f_Theta"]
        s_a = sigma_a * np.sqrt(
            (np.exp(sigma_m**2) - 1) * np.exp(-2 * log_combination) + 1
        )

        m_m = -0.5 * np.log(1 + np.exp(2 * log_combination))
        s_m = np.sqrt(sigma_m**2 - 2 * m_m)

        assert s_m.min() > 0, f"{s_m.min()}, {(sigma_m ** 2 - 2 * m_m).min()}"

        # gather all in one array
        N_pix = forward_map_evals["f_Theta"].shape[0]
        current_bias_std = np.zeros((4, N_pix, self.L))
        current_bias_std[0] = m_a * 1
        current_bias_std[1] = s_a * 1
        current_bias_std[2] = m_m * 1
        current_bias_std[3] = s_m * 1

        return current_bias_std

    def neglog_pdf(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        pixelwise: bool = False,
        full: bool = False,
        idx: Optional[np.ndarray] = None,
    ) -> Union[float, np.ndarray]:

        nlpdf = nll_utils["lambda_"] * np.where(
            nll_utils["censored_mask"], nll_utils["nll_ac"], nll_utils["nll_au"]
        ) + (1 - nll_utils["lambda_"]) * np.where(
            nll_utils["censored_mask"], nll_utils["nll_mc"], nll_utils["nll_mu"]
        )  # (N, L)

        nlpdf = np.nan_to_num(nlpdf)
        nlpdf -= np.log(nll_utils["sigma_a"]) + np.log(nll_utils["sigma_m"])
        # nlpdf /= self.N * self.L
        # nll_utils["censored_mask"].size = self.N * self.L if standard eval
        # nll_utils["censored_mask"].size = N_candidates * self.L if candidates
        if full:
            return nlpdf  # (N, L)

        if pixelwise:
            sum_ = np.sum(nlpdf, axis=1)  # (N,)
            assert sum_.size == forward_map_evals["f_Theta"].shape[0]
            return sum_

        return nlpdf.sum()

    def neglog_pdf_ac(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        omega: np.ndarray,
    ) -> np.ndarray:
        z = (omega - forward_map_evals["f_Theta"] - nll_utils["m_a"]) / nll_utils["s_a"]
        # z = np.nan_to_num(z)

        nll_ac = -log_ndtr(z)
        nll_ac = np.nan_to_num(nll_ac)
        return nll_ac

    def neglog_pdf_au(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        y: np.ndarray,
    ) -> np.ndarray:
        nll_au = cost_au(
            y=y,
            f_Theta=forward_map_evals["f_Theta"],
            m_a=nll_utils["m_a"],
            s_a=nll_utils["s_a"],
        )
        # add a constant to `nll_au` to ensure that it is positive
        # nll_au -= np.log(nll_utils["sigma_a"])
        # nll_au = np.nan_to_num(nll_au)
        return nll_au

    def neglog_pdf_mc(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        log_omega: np.ndarray,
    ) -> np.ndarray:
        z = log_omega - forward_map_evals["log_f_Theta"] - nll_utils["m_m"]
        z /= nll_utils["s_m"]
        # z = np.nan_to_num(z)

        nll_mc = -log_ndtr(z)
        nll_mc = np.nan_to_num(nll_mc)
        return nll_mc

    def neglog_pdf_mu(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        log_y: np.ndarray,
    ) -> np.ndarray:
        nll_mu = cost_mu(
            log_y=log_y,
            log_f_Theta=forward_map_evals["log_f_Theta"],
            m_m=nll_utils["m_m"],
            s_m=nll_utils["s_m"],
        )
        # nll_mu -= np.log(nll_utils["sigma_m"])
        return nll_mu

    def gradient_neglog_pdf(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        """[summary]

        [extended_summary]

        Parameters
        ----------
        x : np.ndarray of shape (N, D)
            [description]
        f_Theta : np.ndarray of shape (N, L), optional
            image of x via forward map, by default None
        grad_f_Theta : np.ndarray of shape (N, D, L), optional
            [description], by default None

        Returns
        -------
        np.ndarray of shape (N, D, L)
            [description]
        """
        grad_ = np.where(
            (self.y == self.omega)[:, None, :],
            # censored
            # grad_nll_ac,
            nll_utils["lambda_"][:, None, :] * nll_utils["grad_nll_ac"]
            + nll_utils["grad_lambda_"] * nll_utils["nll_ac"][:, None, :]
            + (1 - nll_utils["lambda_"])[:, None, :] * nll_utils["grad_nll_mc"]
            - nll_utils["grad_lambda_"] * nll_utils["nll_mc"][:, None, :],
            # uncensored
            nll_utils["lambda_"][:, None, :] * nll_utils["grad_nll_au"]
            + nll_utils["grad_lambda_"] * nll_utils["nll_au"][:, None, :]
            + (1 - nll_utils["lambda_"])[:, None, :] * nll_utils["grad_nll_mu"]
            - nll_utils["grad_lambda_"] * nll_utils["nll_mu"][:, None, :],
        )  # (N, D, L)
        # grad_ = np.where(
        #     np.isfinite(grad_),
        #     grad_,
        #     np.abs(grad_).max(axis=0)[None, :, :],
        # )  # (N, D, L)
        grad_ = np.nan_to_num(grad_)
        return np.sum(grad_, axis=2)  # / (self.N * self.L)

    def gradient_neglog_pdf_ac(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        f_Theta_m_a_omega = forward_map_evals["f_Theta"] + nll_utils["m_a"] - self.omega

        u_1 = utils.norm_pdf_cdf_ratio(-f_Theta_m_a_omega / nll_utils["s_a"])  # (N, L)

        u_2 = (1 / nll_utils["s_a2"])[:, None, :] * (
            (forward_map_evals["grad_f_Theta"] + nll_utils["grad_m_a"])
            * nll_utils["s_a"][:, None, :]
            - f_Theta_m_a_omega[:, None, :] * nll_utils["grad_s_a"]
        )  # (N, D, L)

        grad_ = u_1[:, None, :] * u_2  # (N, D, L)
        # grad_ = np.nan_to_num(grad_)
        # assert np.sum(np.isnan(grad_)) == 0, f"grad_ac : {np.sum(np.isnan(grad_))}"
        return grad_

    def gradient_neglog_pdf_au(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        g_au = gradient_cost_au(
            y=self.y,
            grad_f_Theta=forward_map_evals["grad_f_Theta"],
            f_Theta=forward_map_evals["f_Theta"],
            grad_m_a=nll_utils["grad_m_a"],
            m_a=nll_utils["m_a"],
            grad_s_a2=nll_utils["grad_s_a2"],
            s_a2=nll_utils["s_a2"],
        )
        # g_au = np.nan_to_num(g_au)
        return g_au

    def gradient_neglog_pdf_mc(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        log_f_Theta_m_m_log_omega = (
            forward_map_evals["log_f_Theta"] + nll_utils["m_m"] - np.log(self.omega)
        )

        u_1 = utils.norm_pdf_cdf_ratio(
            -log_f_Theta_m_m_log_omega / nll_utils["s_m"]
        )  # (N, L)
        u_2 = (1 / nll_utils["s_m2"])[:, None, :] * (
            (forward_map_evals["grad_log_f_Theta"] + nll_utils["grad_m_m"])
            * nll_utils["s_m"][:, None, :]
            - log_f_Theta_m_m_log_omega[:, None, :] * nll_utils["grad_s_m"]
        )  # (N, D, L)

        grad_ = u_1[:, None, :] * u_2  # (N, D, L)
        # grad_ = np.nan_to_num(grad_)
        # assert np.sum(np.isnan(grad_)) == 0, f"grad_ac : {np.sum(np.isnan(grad_))}"
        return grad_

    def gradient_neglog_pdf_mu(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        g_mu = gradient_cost_mu(
            log_y=self.log_y,
            grad_log_f_Theta=forward_map_evals["grad_log_f_Theta"],
            log_f_Theta=forward_map_evals["log_f_Theta"],
            grad_m_m=nll_utils["grad_m_m"],
            m_m=nll_utils["m_m"],
            grad_s_m2=nll_utils["grad_s_m2"],
            s_m2=nll_utils["s_m2"],
        )
        # g_mu = np.nan_to_num(g_mu)
        return g_mu

    def hess_diag_neglog_pdf(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        r"""[summary]

        [extended_summary]

        Parameters
        ----------
        x : np.ndarray of shape (N, D)
            [description]
        f_Theta : np.ndarray of shape (N, L), optional
            [description], by default None
        grad_f_Theta : np.ndarray of shape (N, D, L), optional
            [description], by default None
        hess_diag_f_Theta : np.ndarray of shape (N, D, L), optional
            [description], by default None

        Returns
        -------
        np.ndarray of shape (N, D, L)
            [description]
        """
        hess_diag = np.where(
            (self.y == self.omega)[:, None, :],
            # censored
            # hess_diag_ac,
            (
                nll_utils["lambda_"][:, None, :] * nll_utils["hess_diag_ac"]
                + nll_utils["hess_diag_lambda_"] * nll_utils["nll_ac"][:, None, :]
                + 2 * nll_utils["grad_lambda_"] * nll_utils["grad_nll_ac"]
                #
                + (1 - nll_utils["lambda_"])[:, None, :] * nll_utils["hess_diag_mc"]
                - nll_utils["hess_diag_lambda_"] * nll_utils["nll_mc"][:, None, :]
                - 2 * nll_utils["grad_lambda_"] * nll_utils["grad_nll_mc"]
            ),
            # uncensored
            (
                nll_utils["lambda_"][:, None, :] * nll_utils["hess_diag_au"]
                + nll_utils["hess_diag_lambda_"] * nll_utils["nll_au"][:, None, :]
                + 2 * nll_utils["grad_lambda_"] * nll_utils["grad_nll_au"]
                #
                + (1 - nll_utils["lambda_"])[:, None, :] * nll_utils["hess_diag_mu"]
                - nll_utils["hess_diag_lambda_"] * nll_utils["nll_mu"][:, None, :]
                - 2 * nll_utils["grad_lambda_"] * nll_utils["grad_nll_mu"]
            ),
        )  # (N, D, L)
        # hess_diag = np.where(
        #     np.isfinite(hess_diag),
        #     hess_diag,
        #     np.abs(hess_diag).max(axis=0)[None, :, :],
        # )  # (N, D, L)
        hess_diag = np.nan_to_num(hess_diag)
        return np.sum(hess_diag, axis=2)  # / (self.N * self.L)

    def hess_diag_neglog_pdf_ac(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        f_Theta_m_a_omega = forward_map_evals["f_Theta"] + nll_utils["m_a"] - self.omega
        grad_f_Theta_grad_m_a = (
            forward_map_evals["grad_f_Theta"] + nll_utils["grad_m_a"]
        )

        u_1 = utils.norm_pdf_cdf_ratio(-f_Theta_m_a_omega / nll_utils["s_a"])  # (N, L)
        assert u_1.shape == (self.N, self.L)

        u_2 = (1 / nll_utils["s_a2"])[:, None, :] * (
            grad_f_Theta_grad_m_a * nll_utils["s_a"][:, None, :]
            - f_Theta_m_a_omega[:, None, :] * nll_utils["grad_s_a"]
        )  # (N, D, L)
        assert u_2.shape == (self.N, self.D, self.L)

        grad_u_1 = (
            u_2 * (u_1 * (-f_Theta_m_a_omega / nll_utils["s_a"] + u_1))[:, None, :]
        )  # (N, D, L)
        assert grad_u_1.shape == (self.N, self.D, self.L)

        grad_u_2 = (1 / nll_utils["s_a2"] ** 2)[:, None, :] * (
            nll_utils["s_a2"][:, None, :]
            * (
                nll_utils["s_a"][:, None, :]
                * (forward_map_evals["hess_diag_f_Theta"] + nll_utils["hess_diag_m_a"])
                - f_Theta_m_a_omega[:, None, :] * nll_utils["hess_diag_s_a"]
            )
            - nll_utils["grad_s_a2"]
            * (
                nll_utils["s_a"][:, None, :] * grad_f_Theta_grad_m_a
                - f_Theta_m_a_omega[:, None, :] * nll_utils["grad_s_a"]
            )
        )  # (N, D, L)
        assert grad_u_2.shape == (self.N, self.D, self.L)

        hess_diag = grad_u_1 * u_2 + u_1[:, None, :] * grad_u_2
        # hess_diag = np.nan_to_num(hess_diag)
        return hess_diag

    def hess_diag_neglog_pdf_au(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        f_Theta_m_a_y = (
            forward_map_evals["f_Theta"] + nll_utils["m_a"] - self.y
        )  # (N, L)
        grad_f_Theta_grad_m_a = (
            forward_map_evals["grad_f_Theta"] + nll_utils["grad_m_a"]
        )  # (N, D, L)

        u_1 = f_Theta_m_a_y / nll_utils["s_a2"] ** 2  # (N, L)

        grad_u_1 = (1 / nll_utils["s_a2"] ** 4)[:, None, :] * (
            grad_f_Theta_grad_m_a * (nll_utils["s_a2"] ** 2)[:, None, :]
            - 2
            * (f_Theta_m_a_y * nll_utils["s_a2"])[:, None, :]
            * nll_utils["grad_s_a2"]
        )  # (N, D, L)

        u_2 = (
            2 * nll_utils["s_a2"][:, None, :] * grad_f_Theta_grad_m_a
            - f_Theta_m_a_y[:, None, :] * nll_utils["grad_s_a2"]
        )  # (N, D, L)

        grad_u_2 = (
            2
            * nll_utils["s_a2"][:, None, :]
            * (forward_map_evals["hess_diag_f_Theta"] + nll_utils["hess_diag_m_a"])
            #
            + nll_utils["grad_s_a2"] * grad_f_Theta_grad_m_a
            #
            - f_Theta_m_a_y[:, None, :] * nll_utils["hess_diag_s_a2"]
        )  # (N, D, L)

        hess_diag = (
            0.5
            * (
                nll_utils["hess_diag_s_a2"] * nll_utils["s_a2"][:, None, :]
                - nll_utils["grad_s_a2"] ** 2
            )
            / (nll_utils["s_a2"] ** 2)[:, None, :]
        )  # (N, D, L)

        hess_diag += 0.5 * (
            grad_u_1 * u_2 + u_1[:, None, :] * grad_u_2
        )  # hess_diag = np.nan_to_num(hess_diag)
        return hess_diag

    def hess_diag_neglog_pdf_mc(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        log_f_Theta_m_m_log_omega = (
            forward_map_evals["log_f_Theta"] + nll_utils["m_m"] - self.omega
        )
        grad_log_f_Theta_grad_m_m = (
            forward_map_evals["grad_log_f_Theta"] + nll_utils["grad_m_m"]
        )

        u_1 = utils.norm_pdf_cdf_ratio(
            -log_f_Theta_m_m_log_omega / nll_utils["s_m"]
        )  # (N, L)
        assert u_1.shape == (self.N, self.L)

        u_2 = (1 / nll_utils["s_m2"])[:, None, :] * (
            grad_log_f_Theta_grad_m_m * nll_utils["s_m"][:, None, :]
            - log_f_Theta_m_m_log_omega[:, None, :] * nll_utils["grad_s_m"]
        )  # (N, D, L)
        assert u_2.shape == (self.N, self.D, self.L)

        grad_u_1 = (
            u_2
            * (u_1 * (-log_f_Theta_m_m_log_omega / nll_utils["s_m"] + u_1))[:, None, :]
        )  # (N, D, L)
        assert grad_u_1.shape == (self.N, self.D, self.L)

        grad_u_2 = (1 / nll_utils["s_m2"] ** 2)[:, None, :] * (
            nll_utils["s_m2"][:, None, :]
            * (
                nll_utils["s_m"][:, None, :]
                * (forward_map_evals["hess_diag_f_Theta"] + nll_utils["hess_diag_m_m"])
                - log_f_Theta_m_m_log_omega[:, None, :] * nll_utils["hess_diag_s_m"]
            )
            - nll_utils["grad_s_m2"]
            * (
                nll_utils["s_m"][:, None, :] * grad_log_f_Theta_grad_m_m
                - log_f_Theta_m_m_log_omega[:, None, :] * nll_utils["grad_s_m"]
            )
        )  # (N, D, L)
        assert grad_u_2.shape == (self.N, self.D, self.L)

        hess_diag = grad_u_1 * u_2 + u_1[:, None, :] * grad_u_2
        # hess_diag = np.nan_to_num(hess_diag)
        return hess_diag

    def hess_diag_neglog_pdf_mu(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        log_f_Theta_m_m_log_y = (
            forward_map_evals["log_f_Theta"] + nll_utils["m_m"] - self.log_y
        )  # (N, L)
        grad_log_f_Theta_grad_m_m = (
            forward_map_evals["grad_log_f_Theta"] + nll_utils["grad_m_m"]
        )  # (N, D, L)

        u_1 = log_f_Theta_m_m_log_y / nll_utils["s_m2"] ** 2  # (N, L)

        grad_u_1 = (1 / nll_utils["s_m2"] ** 4)[:, None, :] * (
            grad_log_f_Theta_grad_m_m * (nll_utils["s_m2"] ** 2)[:, None, :]
            - 2
            * (log_f_Theta_m_m_log_y * nll_utils["s_m2"])[:, None, :]
            * nll_utils["grad_s_m2"]
        )  # (N, D, L)

        u_2 = (
            2 * nll_utils["s_m2"][:, None, :] * grad_log_f_Theta_grad_m_m
            - log_f_Theta_m_m_log_y[:, None, :] * nll_utils["grad_s_m2"]
        )  # (N, D, L)

        grad_u_2 = 2 * (
            nll_utils["grad_s_m2"] * grad_log_f_Theta_grad_m_m
            + nll_utils["s_m2"][:, None, :]
            * (forward_map_evals["hess_diag_log_f_Theta"] + nll_utils["hess_diag_m_m"])
        ) - (
            nll_utils["grad_s_m2"] * grad_log_f_Theta_grad_m_m
            + log_f_Theta_m_m_log_y[:, None, :] * nll_utils["hess_diag_s_m2"]
        )  # (N, D, L)

        hess_diag = (
            0.5
            * (
                nll_utils["hess_diag_s_m2"] * nll_utils["s_m2"][:, None, :]
                - nll_utils["grad_s_m2"] ** 2
            )
            / (nll_utils["s_m2"] ** 2)[:, None, :]
        )  # (N, D, L)

        hess_diag += 0.5 * (
            grad_u_1 * u_2 + u_1[:, None, :] * grad_u_2
        )  # hess_diag = np.nan_to_num(hess_diag)
        return hess_diag

    def evaluate_all_nll_utils(
        self,
        forward_map_evals: dict,
        idx: Optional[np.ndarray] = None,
        compute_derivatives: bool = True,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict:
        assert (
            np.sum(np.isnan(forward_map_evals["log_f_Theta"])) == 0
        ), f"before entering bias and std : {np.sum(np.isnan(forward_map_evals['log_f_Theta']))}"

        nll_utils = {}
        # * bias and variance
        if idx is None:
            N_pix = self.N * 1
            sigma_a = self.sigma_a * 1
            sigma_m = self.sigma_m * 1
            y = self.y * 1
            log_y = self.log_y * 1
            omega = self.omega * 1
        else:
            n_pix = idx.size
            k_mtm = forward_map_evals["f_Theta"].shape[0] // n_pix
            N_pix = forward_map_evals["f_Theta"].shape[0]
            assert n_pix * k_mtm == N_pix

            sigma_a = np.zeros((n_pix, k_mtm, self.L))
            sigma_m = np.zeros((n_pix, k_mtm, self.L))
            y = np.zeros((n_pix, k_mtm, self.L))
            omega = np.zeros((n_pix, k_mtm, self.L))

            for i_pix in range(n_pix):
                sigma_a[i_pix, :, :] = self.sigma_a[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                sigma_m[i_pix, :, :] = self.sigma_m[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                y[i_pix, :, :] = self.y[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                omega[i_pix, :, :] = self.omega[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )

            sigma_a = (
                sigma_a.transpose((2, 0, 1)).reshape((self.L, N_pix)).T
            )  # .reshape((N_pix, self.L))
            sigma_m = (
                sigma_m.transpose((2, 0, 1)).reshape((self.L, N_pix)).T
            )  # .reshape((N_pix, self.L))
            y = (
                y.transpose((2, 0, 1)).reshape((self.L, N_pix)).T
            )  # .reshape((N_pix, self.L))
            log_y = np.log(y)
            omega = (
                omega.transpose((2, 0, 1)).reshape((self.L, N_pix)).T
            )  # .reshape((N_pix, self.L))            omega

        # * -----
        nll_utils["censored_mask"] = (y <= omega) * 1
        nll_utils["sigma_a"] = sigma_a * 1
        nll_utils["sigma_m"] = sigma_m * 1

        sigma_m2 = sigma_m**2
        log_combination = (
            np.log(sigma_a) - forward_map_evals["log_f_Theta"] - sigma_m2 / 2
        )
        exp_m2_log_combin = np.exp(-2 * log_combination)
        exp_sigma_m_squared = np.exp(sigma_m2)
        # exp_sigma_m_squared_div2_m1 = np.exp(sigma_m2 / 2) - 1  # (N, L)

        # * computation of bias and variances
        nll_utils["m_a"] = np.zeros_like(y)
        nll_utils["s_a"] = sigma_a * np.sqrt(
            (exp_sigma_m_squared - 1)
            * np.exp(2 * (forward_map_evals["log_f_Theta"] - np.log(sigma_a)))
            + 1
        )
        nll_utils["s_a2"] = nll_utils["s_a"] ** 2

        nll_utils["m_m"] = -0.5 * (sigma_m2 + np.log(1 + 1 / exp_m2_log_combin))

        nll_utils["s_m2"] = -2 * nll_utils["m_m"]
        nll_utils["s_m"] = np.sqrt(nll_utils["s_m2"])

        assert (
            nll_utils["s_m"].min() > 0
        ), f"{nll_utils['s_m'].min()}, {(sigma_m2 - 2 * nll_utils['m_m']).min()}"

        if compute_derivatives:
            nll_utils["grad_m_a"] = np.zeros((N_pix, self.D, self.L))

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_m_a"] = np.zeros((N_pix, self.D, self.L))

            nll_utils["grad_s_a2"] = (
                2
                * (forward_map_evals["f_Theta"] * (exp_sigma_m_squared - 1))[:, None, :]
                * forward_map_evals["grad_f_Theta"]
            )  # (N, D, L)

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_s_a2"] = (
                    2
                    * (exp_sigma_m_squared - 1)[:, None, :]
                    * (
                        forward_map_evals["f_Theta"][:, None, :]
                        * forward_map_evals["hess_diag_f_Theta"]
                        + forward_map_evals["grad_f_Theta"] ** 2
                    )
                )  # (N, D, L)

            nll_utils["grad_s_a"] = (
                1 / (2 * nll_utils["s_a"])[:, None, :]
            ) * nll_utils["grad_s_a2"]

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_s_a"] = (1 / (2 * nll_utils["s_a2"]))[
                    :, None, :
                ] * (
                    nll_utils["hess_diag_s_a2"] * nll_utils["s_a"][:, None, :]
                    - nll_utils["grad_s_a2"] * nll_utils["grad_s_a"]
                )  # (N, D, L)

            nll_utils["grad_m_m"] = (
                1 / (1 + exp_m2_log_combin) / forward_map_evals["f_Theta"]
            )[:, None, :] * forward_map_evals[
                "grad_f_Theta"
            ]  # (N, D, L)

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_m_m"] = (
                    1 / (forward_map_evals["f_Theta"] * (1 + exp_m2_log_combin)) ** 2
                )[:, None, :] * (
                    forward_map_evals["hess_diag_f_Theta"]
                    * (forward_map_evals["f_Theta"] * (1 + exp_m2_log_combin))[
                        :, None, :
                    ]
                    - forward_map_evals["grad_f_Theta"] ** 2
                    * (1 + 3 * exp_m2_log_combin)[:, None, :]
                )  # (N, D, L)

            nll_utils["grad_s_m2"] = -2 * nll_utils["grad_m_m"]  # (N, D, L)

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_s_m2"] = (
                    -2 * nll_utils["hess_diag_m_m"]
                )  # (N, D, L)

            nll_utils["grad_s_m"] = (1 / (2 * nll_utils["s_m"]))[
                :, None, :
            ] * nll_utils[
                "grad_s_m2"
            ]  # (N, D, L)

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_s_m"] = (1 / (2 * nll_utils["s_m2"]))[
                    :, None, :
                ] * (
                    nll_utils["hess_diag_s_m2"] * nll_utils["s_m"][:, None, :]
                    - nll_utils["grad_s_m"] * nll_utils["grad_s_m2"]
                )  # (N, D, L)

            # assert np.sum(np.isnan(forward_map_evals["f_Theta"])) == 0.0
            # assert (
            #     forward_map_evals["f_Theta"].min() > 0.0
            # ), f"{forward_map_evals['f_Theta'].min()}, {forward_map_evals['log_f_Theta'].min()}"
            # assert np.sum(np.isnan(forward_map_evals["grad_f_Theta"])) == 0.0
            # assert np.sum(np.isnan(forward_map_evals["hess_diag_f_Theta"])) == 0.0

            # assert np.sum(np.isnan(forward_map_evals["log_f_Theta"])) == 0.0
            # assert np.sum(np.isnan(forward_map_evals["grad_log_f_Theta"])) == 0.0
            # assert np.sum(np.isnan(forward_map_evals["hess_diag_log_f_Theta"])) == 0.0

            # assert np.sum(np.isnan(m_a)) == 0.0
            # assert np.sum(np.isnan(grad_m_a)) == 0.0
            # assert np.sum(np.isnan(hess_diag_m_a)) == 0.0
            # assert m_a.shape == (self.N, self.L)
            # assert grad_m_a.shape == (self.N, self.D, self.L)
            # assert hess_diag_m_a.shape == (self.N, self.D, self.L)

            # assert np.sum(np.isnan(s_a2)) == 0.0
            # assert np.all(s_a2 > 0)
            # assert np.sum(np.isnan(grad_s_a2)) == 0.0
            # assert np.sum(np.isnan(hess_diag_s_a2)) == 0.0
            # assert s_a2.shape == (self.N, self.L)
            # assert grad_s_a2.shape == (self.N, self.D, self.L)
            # assert hess_diag_s_a2.shape == (self.N, self.D, self.L)

            # assert np.sum(np.isnan(s_a)) == 0.0
            # assert np.all(s_a > 0)
            # assert np.sum(np.isnan(grad_s_a)) == 0.0
            # assert np.sum(np.isnan(hess_diag_s_a)) == 0.0
            # assert s_a.shape == (self.N, self.L)
            # assert grad_s_a.shape == (self.N, self.D, self.L)
            # assert hess_diag_s_a.shape == (self.N, self.D, self.L)

            # assert np.sum(np.isnan(m_m)) == 0.0
            # assert np.sum(np.isnan(grad_m_m)) == 0.0
            # assert np.sum(np.isnan(hess_diag_m_m)) == 0.0
            # assert m_m.shape == (self.N, self.L)
            # assert grad_m_m.shape == (self.N, self.D, self.L)
            # assert hess_diag_m_m.shape == (self.N, self.D, self.L)

            # assert np.sum(np.isnan(s_m2)) == 0.0
            # assert np.sum(np.isnan(grad_s_m2)) == 0.0
            # assert np.sum(np.isnan(hess_diag_s_m2)) == 0.0
            # assert s_m2.shape == (self.N, self.L)
            # assert grad_s_m2.shape == (self.N, self.D, self.L)
            # assert hess_diag_s_m2.shape == (self.N, self.D, self.L)

            # assert np.sum(np.isnan(s_m)) == 0.0
            # assert np.sum(np.isnan(grad_s_m)) == 0.0
            # assert np.sum(np.isnan(hess_diag_s_m)) == 0.0
            # assert s_m.shape == (self.N, self.L)
            # assert grad_s_m.shape == (self.N, self.D, self.L)
            # assert hess_diag_s_m.shape == (self.N, self.D, self.L)

        # * mixture weight
        nll_utils["lambda_"] = self.model_mixing_param(forward_map_evals, idx)

        if compute_derivatives:
            nll_utils["grad_lambda_"] = self.grad_model_mixing_param(forward_map_evals)
            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_lambda_"] = self.hess_diag_model_mixing_param(
                    forward_map_evals
                )

        nll_utils["nll_au"] = self.neglog_pdf_au(forward_map_evals, nll_utils, y)
        nll_utils["nll_ac"] = self.neglog_pdf_ac(forward_map_evals, nll_utils, y)
        nll_utils["nll_mu"] = self.neglog_pdf_mu(forward_map_evals, nll_utils, log_y)
        nll_utils["nll_mc"] = self.neglog_pdf_mc(
            forward_map_evals, nll_utils, np.log(omega)
        )

        if compute_derivatives:
            nll_utils["grad_nll_au"] = self.gradient_neglog_pdf_au(
                forward_map_evals, nll_utils
            )
            nll_utils["grad_nll_mu"] = self.gradient_neglog_pdf_mu(
                forward_map_evals, nll_utils
            )
            nll_utils["grad_nll_ac"] = self.gradient_neglog_pdf_ac(
                forward_map_evals, nll_utils
            )
            nll_utils["grad_nll_mc"] = self.gradient_neglog_pdf_mc(
                forward_map_evals, nll_utils
            )

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_ac"] = self.hess_diag_neglog_pdf_ac(
                    forward_map_evals, nll_utils
                )
                nll_utils["hess_diag_au"] = self.hess_diag_neglog_pdf_au(
                    forward_map_evals, nll_utils
                )
                nll_utils["hess_diag_mc"] = self.hess_diag_neglog_pdf_mc(
                    forward_map_evals, nll_utils
                )
                nll_utils["hess_diag_mu"] = self.hess_diag_neglog_pdf_mu(
                    forward_map_evals, nll_utils
                )

        return nll_utils



from typing import Optional, Union

import numba
import numpy as np

from beetroots.modelling.priors.abstract_spatial_prior import SpatialPrior


@numba.njit()
def compute_laplacian(
    Theta: np.ndarray, list_edges: np.ndarray, idx_pix: np.ndarray
) -> np.ndarray:
    r"""evaluates the image Laplacian for each of the D maps, :math:`\Delta \Theta_{\cdot d}`

    Parameters
    ----------
    Theta : np.ndarray of shape (N, D)
        D vectors of N pixels
    list_edges : np.ndarray
        set of edges in the graph induced by the spatial regularization

    Returns
    -------
    np.ndarray of shape (n_pix, D)
        image Laplacian for each of the D maps
    """
    laplacian_ = np.zeros((idx_pix.size, *Theta.shape[1:]))

    for i, n in enumerate(idx_pix):
        mask_i_m = list_edges[:, 1] == n
        mask_i_p = list_edges[:, 0] == n

        laplacian_[i] -= np.sum(
            (Theta[list_edges[mask_i_p, 1], :] - Theta[list_edges[mask_i_p, 0], :]),
            axis=0,
        )
        laplacian_[i] += np.sum(
            (Theta[list_edges[mask_i_m, 1], :] - Theta[list_edges[mask_i_m, 0], :]),
            axis=0,
        )
    return laplacian_  # (N, D)


@numba.njit()
def compute_hadamard_discrete_gradient(
    Theta: np.ndarray, list_edges: np.ndarray, idx_pix: np.ndarray
) -> np.ndarray:
    r"""evaluates the sum of squared differences for each of the D maps, :math:`\Delta \Theta_{\cdot d}`

    Parameters
    ----------
    Theta : np.ndarray of shape (N, D)
        D vectors of N pixels
    list_edges : np.ndarray
        set of edges in the graph induced by the spatial regularization
    idx_pix : np.ndarray
        array of the indices of the pixels to consider

    Returns
    -------
    np.ndarray of shape (n_pix, D)
        image of  for each of the D maps
    """
    hadamard_gradient_ = np.zeros((idx_pix.size, *Theta.shape[1:]))

    for i, n in enumerate(idx_pix):
        mask_i_m = list_edges[:, 1] == n
        mask_i_p = list_edges[:, 0] == n

        hadamard_gradient_[i] += np.sum(
            (Theta[list_edges[mask_i_p, 1], :] - Theta[list_edges[mask_i_p, 0], :])
            ** 2,
            axis=0,
        )
        hadamard_gradient_[i] += np.sum(
            (Theta[list_edges[mask_i_m, 1], :] - Theta[list_edges[mask_i_m, 0], :])
            ** 2,
            axis=0,
        )
    return hadamard_gradient_  # (N, D)


class L22DiscreteGradSpatialPrior(SpatialPrior):
    r"""L22 smooth spatial prior, valid for both 1D and 2D tensors. Its pdf is defined as

    .. math::

        \forall d \in [1, D], \quad \pi(\Theta_{\cdot d}) \propto \exp \left[- \tau_d \Vert D \Theta_{\cdot d} \Vert_F^2 \right] = \exp \left[- \frac{\tau_d}{2} \sum_{n \in \mathcal{N}} \sum_{i\in V_n}\left( \Theta_{n, d} - \Theta_{i, d} \right)^2 \right]

    where  :math:`\Vert \cdot \Vert_F` denotes the Fröbenius norm and :math:`D\Theta_{\cdot d}` is the discrete gradient operator for vector :math:\Theta_{\cdot d}`.

    It does handle the MTM case with an additional dimension (thourgh the `full` dimension).
    """

    def neglog_pdf(
        self,
        Theta: np.ndarray,
        idx_pix: np.ndarray,
        with_weights: bool = True,
        full: bool = False,
        pixelwise: bool = False,
        chromatic_gibbs: Union[bool, str] = False,
    ) -> np.ndarray:
        assert np.sum([pixelwise, full]) < 2
        if idx_pix.size < self.N:
            chromatic_gibbs = True
        factor = 1 if chromatic_gibbs else 1 / 2

        if self.list_edges.size > 0:
            hadamard_gradient_ = compute_hadamard_discrete_gradient(
                Theta, self.list_edges, idx_pix
            )
            neglog_p = hadamard_gradient_
        else:
            neglog_p = np.zeros((idx_pix.size, *Theta.shape[1:]))

        if with_weights:
            neglog_p = self.weights * neglog_p

        if pixelwise:
            neglog_p = np.sum(neglog_p, axis=1)
        elif not full:
            neglog_p = np.sum(neglog_p, axis=0)

        if isinstance(chromatic_gibbs, bool):
            return factor * neglog_p  # (D,) if not pixelwise or (N, D) if pixelwise
        elif chromatic_gibbs == "both":
            return 1 * neglog_p, 1 / 2 * neglog_p
        else:
            raise ValueError(
                'wrong value for "chromatic_gibbs", should be a boolean or the string "both"'
            )

    def gradient_neglog_pdf(self, Theta: np.ndarray, idx_pix: np.ndarray) -> np.ndarray:
        assert Theta.shape == (self.N, self.D)

        laplacian_ = compute_laplacian(Theta, self.list_edges, idx_pix)

        return self.weights[None, :] * 2 * laplacian_  # (N, D)

    def hess_diag_neglog_pdf(
        self, Theta: np.ndarray, idx_pix: np.ndarray
    ) -> np.ndarray:
        hess_diag = np.zeros((idx_pix.size, *Theta.shape[1:]), dtype=np.float64)

        if self.list_edges.size > 0:
            idx, counts = np.unique(self.list_edges.flatten(), return_counts=True)
            indices_kept_1 = np.isin(idx_pix, idx, assume_unique=True)
            indices_kept_2 = np.isin(idx, idx_pix, assume_unique=True)

            hess_diag[indices_kept_1, :] += counts[indices_kept_2, None] * np.ones(
                (indices_kept_1.sum(), self.D)
            )

        # hess_diag /= self.N * self.D
        return self.weights[None, :] * 2 * hess_diag  # (N, D)



r"""Contains a class that defines the transition between the sampling scale and user friendly / interpretable scale
"""
from typing import List

import numba
import numpy as np

from beetroots.space_transform.abstract_transform import Scaler


@numba.njit()
def _from_scaled_to_lin(
    Theta_scaled: np.ndarray,
    mean_: np.ndarray,
    std_: np.ndarray,
    list_is_log: List[bool],
    D: int,
    LOG_10: float,
) -> np.ndarray:
    Theta_linscale = np.zeros_like(Theta_scaled)

    for d in range(D):
        rescaled = std_[d] * Theta_scaled[:, d] + mean_[d]
        if list_is_log[d]:
            Theta_linscale[:, d] = np.exp(LOG_10 * rescaled)
        else:
            Theta_linscale[:, d] = rescaled * 1

    return Theta_linscale


@numba.njit()
def _from_lin_to_scaled(
    Theta_linscale: np.ndarray,
    mean_: np.ndarray,
    std_: np.ndarray,
    list_is_log: List[bool],
    D: int,
) -> np.ndarray:
    Theta_scaled = np.zeros_like(Theta_linscale)

    for d in range(D):
        if list_is_log[d]:
            scaled = np.log10(Theta_linscale[:, d])
        else:
            scaled = Theta_linscale[:, d] * 1

        Theta_scaled[:, d] = (scaled - mean_[d]) / std_[d]

    return Theta_scaled


class MyScaler(Scaler):
    r"""Defines the scale used during sampling and the transforms to navigate from one scale to the other. The transformation is a normalization (defined with a mean `mean_` and standard deviation `std_`) for each physical parameter, defined on the log10 scale or on the linear scale depending on `list_is_log`.

    .. note::

        If one of the physical parameters is the scaling factor :math:`\kappa`, its mean is set to 0 and its std to 1 / np.log(10), so that it is not normalized regardless to its sampling scale (log10 or linear).

        The std = 1 / np.log(10) for kappa yields a scaled validity interval that is about [-2.7, 2.7] (for a [0.1, 10] true validity interval), ie comparable to that of other normalized parameters.
    """

    __slots__ = ("D", "mean_", "std_", "list_is_log")
    LOG_10 = np.log(10.0)

    def __init__(
        self,
        mean_: np.ndarray,
        std_: np.ndarray,
        list_is_log: List[bool],
    ):
        assert mean_.shape == std_.shape
        assert mean_.size in [len(list_is_log), len(list_is_log) - 1]

        # if there is a kappa in the set of parameters
        # (if there is no kappa, then each parameter should have an associated
        # mean and std in the scaler)
        if mean_.size == len(list_is_log) - 1:
            # kappa: mean = 0 and std = 1 / np.log(10)
            mean_ = np.array([0.0] + list(mean_))
            std_ = np.array([1.0 / self.LOG_10] + list(std_))

        self.D = mean_.size
        r"""int: total number of physical parameters that require a standard scaler, including the scaling factor :math:`\kappa`"""

        self.mean_ = mean_
        r"""np.ndarray of shape (D,): mean of the D components :math:`\theta_d`, used in the data normalization"""

        self.std_ = std_
        r"""np.ndarray of shape (D,): standard deviation of the D components :math:`\theta_d`, used in the data normalization"""

        self.list_is_log = list_is_log
        r"""list of bool of length D: whether the normalization should be applied on the log10 scale or in the linear scale"""

    def from_scaled_to_lin(self, Theta_scaled: np.ndarray) -> np.ndarray:
        assert len(Theta_scaled.shape) == 2, Theta_scaled.shape
        assert Theta_scaled.shape[1] == self.D, Theta_scaled.shape

        Theta_linscale = _from_scaled_to_lin(
            Theta_scaled,
            self.mean_,
            self.std_,
            self.list_is_log,
            self.D,
            self.LOG_10,
        )
        return Theta_linscale

    def from_lin_to_scaled(self, Theta_linscale: np.ndarray) -> np.ndarray:
        assert len(Theta_linscale.shape) == 2, Theta_linscale.shape
        assert Theta_linscale.shape[1] == self.D, Theta_linscale.shape

        Theta_scaled = _from_lin_to_scaled(
            Theta_linscale,
            self.mean_,
            self.std_,
            self.list_is_log,
            self.D,
        )
        return Theta_scaled



import pickle
from typing import Dict, Optional, Tuple

import numpy as np

from beetroots.approx_optim.forward_map.abstract_forward_map import (
    ApproxOptimForwardMap,
)
from beetroots.modelling.forward_maps.regression_poly import PolynomialApprox
from beetroots.space_transform.transform import MyScaler


class ApproxOptimPolynomialReg(ApproxOptimForwardMap):
    r"""handles the generation of a dataset of :math:`\log_{10} f_{\ell}(\theta)` values  for a polynomial forward map

    .. warning::

        Unfinished class.
        Needs to be updated for D_sampling to remove the angle parameter.
        Needs work on the :class:`.PolynomialApprox` class
    """

    def setup_forward_map(
        self,
        forward_model_name: str,
        dict_fixed_params: Dict[str, Optional[float]],
        dict_is_log_scale_params: Dict[str, bool],
    ) -> Tuple[MyScaler, PolynomialApprox]:

        with open(
            f"{self.MODELS_PATH}/{forward_model_name}/scaler.pickle",
            "rb",
        ) as file_:
            scaler_sklearn = pickle.load(file_)

        scaler = MyScaler(
            mean_=scaler_sklearn.mean_[:-1].flatten(),
            std_=scaler_sklearn.scale_[:-1].flatten(),
            list_is_log=list(dict_is_log_scale_params.values()),
        )

        # transformation from linear scale (in degrees) to scaled
        angle = dict_fixed_params["angle"]
        assert angle is not None
        angle_scaled = (angle - 30.0) / 20.0

        dict_fixed_params_scaled = self.scale_dict_fixed_params(
            scaler, dict_fixed_params
        )

        # load forward model
        forward_map = PolynomialApprox(
            self.MODELS_PATH,
            forward_model_name,
            dict_fixed_params_scaled,
            angle_scaled,
        )
        forward_map.restrict_to_output_subset(self.list_lines)
        return scaler, forward_map

    def compute_log10_f_Theta(
        self,
        dict_forward_model: dict,
        lower_bounds_lin: np.ndarray,
        upper_bounds_lin: np.ndarray,
    ):
        r""".. warning::

        Unfinished method.
        Needs to be updated for D_sampling to remove the angle parameter.
        Needs work on the :class:`.PolynomialApprox` class
        """
        scaler, forward_map = self.setup_forward_map(**dict_forward_model)

        lower_bounds = scaler.from_lin_to_scaled(
            lower_bounds_lin.reshape((1, self.D)),
        ).flatten()
        upper_bounds = scaler.from_lin_to_scaled(
            upper_bounds_lin.reshape((1, self.D)),
        ).flatten()
        Theta = self.sample_theta(lower_bounds, upper_bounds)

        # the division is to get log in base 10
        log10_f_Theta = forward_map.evaluate_log(Theta) / np.log(10)
        assert log10_f_Theta.shape == (self.N_samples_theta, self.L)

        return log10_f_Theta



import os
import sys

import numpy as np
import yaml

from beetroots.approx_optim.nn_bo import ApproxParamsOptimNNBO

if __name__ == "__main__":
    yaml_file, path_data, path_models, path_outputs = ApproxParamsOptimNNBO.parse_args()

    # step 1: read ``.yaml`` input file
    with open(os.path.abspath(f"{path_data}/{yaml_file}")) as f:
        params = yaml.safe_load(f)

    # step 2: run the Bayesian optimization
    approx_optim = ApproxParamsOptimNNBO(
        params["list_lines"],
        sigma_a_raw=params["sigma_a"],
        sigma_m=np.log(params["sigma_m_float_linscale"]),
        **params["simu_init"],
        path_models=path_models,
        path_outputs=path_outputs,
    )

    approx_optim.main(**params["main_params"])

    # approx_optim.save_results_in_data_folder(
    #     path_data, params['filename_err']
    # )



import abc
from typing import Dict

import matplotlib.pyplot as plt
from matplotlib import patches

from beetroots.inversion.plots.map_shaper import MapShaper


class AbstractPlots2D(abc.ABC):
    """abstract class for drawing figures of maps of observations, noise standard deviations, estimated physical parameters, estimated credibility interval sizes, etc."""

    def __init__(
        self,
        map_shaper: MapShaper,
        pixels_of_interest: Dict[int, str] = {},
    ):
        self.map_shaper = map_shaper
        r"""MapShaper: defines the transformation from vectors to 2D maps"""

        self.pixels_of_interest_names = pixels_of_interest
        r"""dict: (coordinate, name) pair of some user-informed pixels to be highlighted. These pixels will be outlines with a black square in figures."""

        coords = map_shaper.from_vector_idx_to_map_coords(
            list(pixels_of_interest.keys())
        )
        self.pixels_of_interest_coords = coords
        r"""list: coordinates of some user-informed pixels to be highlighted. These pixels will be outlines with a black square in figures."""

    def _draw_rect_on_pixels_of_interest(self):
        """Draws a rectangle around the pixels of interest. Requires the class to have a `pixels_of_interest_coords` attribute, as a Dict with the vector index as keys and the corresponding map coordinates as values."""
        ax = plt.gca()

        for _, coords in self.pixels_of_interest_coords.items():
            rect = patches.Rectangle(
                (coords[0] - 0.5, coords[1] - 0.5),
                1,
                1,
                linewidth=1,
                edgecolor="k",
                facecolor="none",
            )
            ax.add_patch(rect)

        return



import abc
from typing import Dict, Optional, Tuple

import numpy as np

from beetroots.space_transform.abstract_transform import Scaler


class ApproxOptimForwardMap(abc.ABC):
    r"""abstract class that handles the generation of a dataset of :math:`\log_{10} f_{\ell}(\theta)` values"""

    @abc.abstractmethod
    def compute_log10_f_Theta(self):
        r"""Evaluates :math:`\log_{10} f` on a list of ``self.N_samples_theta`` :math:`\theta` values, each randomly generated in the hypercube defined by its lower and upper bounds.

        Parameters
        ----------
        forward_model_name : str
            name of the forward model to load (i.e., of the corresponding folder)
        angle : float
            angle at which the cloud is observed
        lower_bounds_lin : np.ndarray of shape (D,)
            array of lower bounds in linear scale
        upper_bounds_lin : np.ndarray of shape (D,)
            array of upper bounds in linear scale

        Returns
        -------
        np.ndarray of shape (self.N_samples_theta, L)
            evaluations of the ``self.N_samples_theta`` :math:`\theta` values for the L considered lines.
        """
        pass

    def scale_dict_fixed_params(
        self, scaler: Scaler, dict_fixed_params: Dict[str, Optional[float]]
    ) -> Dict[str, Optional[float]]:
        r"""transforms the fixed parameters from their value in their natural spae to their value in the space in which they are to be sampled.

        Note
        ----
        since the scaler does not accept None or np.NaN, this transformation really needed a dedicated (though simple) method

        Parameters
        ----------
        scaler : Scaler
            contains the transformation of the physical parameters values from their natural space to their scaled space (in which the sampling happens) and its inverse
        dict_fixed_params : Dict[str, Optional[float]]
            contains the value of the fixed parameters in the physical parameter natural space. For example, {"kappa":None, "Pth":None, "G0":None, "AV":None, "angle":0.} (the values with None indicate that the parameter is not fixed, so here, only the angle is set, to 0 degree).

        Returns
        -------
        Dict[str, Optional[float]]
            contains the value of the fixed parameters in the space in which the physical parameter are to be sampled
        """
        arr_fixed_values = np.ones((1, len(dict_fixed_params)))
        for i, name in enumerate(dict_fixed_params.keys()):
            if dict_fixed_params[name] is not None:
                arr_fixed_values[0, i] = dict_fixed_params[name]

        arr_fixed_values_scaled = scaler.from_lin_to_scaled(arr_fixed_values)

        dict_fixed_params_scaled = {
            name: arr_fixed_values_scaled[0, i]
            if dict_fixed_params[name] is not None
            else None
            for i, name in enumerate(dict_fixed_params.keys())
        }
        return dict_fixed_params_scaled



import pickle
from typing import Dict, Optional, Tuple

import numpy as np

from beetroots.approx_optim.forward_map.abstract_forward_map import (
    ApproxOptimForwardMap,
)
from beetroots.modelling.forward_maps.neural_network_approx import NeuralNetworkApprox
from beetroots.space_transform.transform import MyScaler


class ApproxOptimNN(ApproxOptimForwardMap):
    r"""handles the generation of a dataset of :math:`\log_{10} f_{\ell}(\theta)` values for a neural network forward map"""

    def setup_forward_map(
        self,
        forward_model_name: str,
        force_use_cpu: bool,
        fixed_params: Dict[str, Optional[float]],
        is_log_scale_params: Dict[str, bool],
    ) -> Tuple[MyScaler, NeuralNetworkApprox]:
        print(f"fixed values: {fixed_params}")

        with open(
            f"{self.MODELS_PATH}/{forward_model_name}/scaler.pickle",
            "rb",
        ) as file_:
            scaler_sklearn = pickle.load(file_)

        scaler = MyScaler(
            mean_=scaler_sklearn.mean_.flatten(),
            std_=scaler_sklearn.scale_.flatten(),
            list_is_log=list(is_log_scale_params.values()),
        )

        # transformation from linear scale (in degrees) to scaled

        # angle_scaled = (angle - 30.0) / 20.0

        # eg {"kappa":None, "Pth":None, "G0":None, "AV":None, "angle":0.}
        fixed_params_scaled = self.scale_dict_fixed_params(scaler, fixed_params)

        # load forward model
        forward_map = NeuralNetworkApprox(
            self.MODELS_PATH,
            forward_model_name,
            fixed_params_scaled,
            device="cpu" if force_use_cpu else None,
        )
        forward_map.restrict_to_output_subset(self.list_lines)
        return scaler, forward_map

    def compute_log10_f_Theta(
        self,
        dict_forward_model: dict,
        lower_bounds_lin: np.ndarray,
        upper_bounds_lin: np.ndarray,
    ) -> np.ndarray:

        scaler, forward_map = self.setup_forward_map(**dict_forward_model)

        lower_bounds = scaler.from_lin_to_scaled(
            lower_bounds_lin.reshape((1, self.D)),
        ).flatten()  # (D,)
        upper_bounds = scaler.from_lin_to_scaled(
            upper_bounds_lin.reshape((1, self.D)),
        ).flatten()  # (D,)

        # restrict bounds to the unfixed parameters
        lower_bounds = lower_bounds[self.list_idx_sampling] * 1
        upper_bounds = upper_bounds[self.list_idx_sampling] * 1

        fixed_params_scaled = self.scale_dict_fixed_params(
            scaler, dict_forward_model["fixed_params"]
        )
        print(fixed_params_scaled)

        Theta = np.zeros((self.K**self.D_sampling, self.D))
        Theta[:, self.list_idx_sampling] = self.sample_theta(lower_bounds, upper_bounds)

        for i, v in enumerate(fixed_params_scaled.values()):
            if v is not None:
                assert np.allclose(Theta[:, i], 0.0), Theta[:10]
                Theta[:, i] = v * 1

        print(f"Theta.shape = {Theta.shape}")

        # the division is to get log in base 10
        print("starting computation of set of log10 f(theta) values")

        log10_f_Theta = forward_map.evaluate_log(Theta) / np.log(10)
        assert log10_f_Theta.shape == (
            self.N_samples_theta,
            self.L,
        ), f"log10_f_Theta.shape = {log10_f_Theta.shape}, (self.N_samples_theta, self.L) = {(self.N_samples_theta, self.L)}"

        print(f"log10_f_Theta.shape = {log10_f_Theta.shape}")
        print("computation of set of log10 f(theta) values done")

        return log10_f_Theta



"""(deprecated) Implementation of L22 Laplacian prior. False implementation."""

import warnings
from typing import Optional

import numba
import numpy as np
import pandas as pd
from numba.typed import Dict, List

from beetroots.modelling.priors.abstract_spatial_prior import SpatialPrior


@numba.njit()
def compute_laplacian(
    Theta: np.ndarray, list_edges: np.ndarray, idx_pix: np.ndarray
) -> np.ndarray:
    r"""evaluates the image Laplacian for each of the D maps, :math:`\Delta \Theta_{\cdot d}`

    Parameters
    ----------
    Theta : np.ndarray of shape (N, D)
        D vectors of N pixels
    list_edges : np.ndarray
        set of edges in the graph induced by the spatial regularization

    Returns
    -------
    np.ndarray of shape (N, D)
        image Laplacian for each of the D maps
    """
    laplacian_ = np.zeros((idx_pix.size, *Theta.shape[1:]))

    for i, n in enumerate(idx_pix):
        mask_i_m = list_edges[:, 1] == n
        mask_i_p = list_edges[:, 0] == n

        laplacian_[i] -= np.sum(
            (Theta[list_edges[mask_i_p, 1], :] - Theta[list_edges[mask_i_p, 0], :]),
            axis=0,
        )
        laplacian_[i] += np.sum(
            (Theta[list_edges[mask_i_m, 1], :] - Theta[list_edges[mask_i_m, 0], :]),
            axis=0,
        )
    return laplacian_  # (N, D)


@numba.njit()
def compute_neglog_chromatic_from_laplacian2(
    laplacian_2: np.ndarray, list_edges: np.ndarray, idx_pix: np.ndarray
) -> np.ndarray:
    r"""evaluates the negative log pdf for each of the D maps, :math:`\Delta \Theta_{\cdot d}`

    Parameters
    ----------
    Theta : np.ndarray of shape (N, D)
        D vectors of N pixels
    dict_neighbors : np.ndarray
        set of neighbors for each pixel
    weights : np.ndarray
        weights for each pixel

    Returns
    -------
    np.ndarray of shape (N, D)
        negative log pdf for each of the D maps
    """
    neglog_p = np.zeros(
        (idx_pix.size, *laplacian_2.shape[1:])
    )  # (n_pix, D) or (n_pix, k_mtm, D)

    for i, n in enumerate(idx_pix):
        mask_i_m = list_edges[:, 1] == n
        mask_i_p = list_edges[:, 0] == n
        neighbors_indices = np.asarray(
            list(list_edges[mask_i_m, 0]) + list(list_edges[mask_i_p, 1]),
            dtype=np.int64,
        )

        neglog_p[i] += laplacian_2[n]
        neglog_p[i] += np.sum(laplacian_2[neighbors_indices, :], axis=0)

    return neglog_p


@numba.njit()
def compute_gradient_from_laplacian(
    laplacian_: np.ndarray, idx_pix: np.ndarray, list_edges: np.ndarray
) -> np.ndarray:
    """evaluates the gradient from the Laplacian matrix

    Parameters
    ----------
    laplacian_ : np.ndarray
        Laplacian matrix
    list_edges : np.ndarray
        array containing all the graph edges, identified with pairs of indices of neighboring pixels

    Returns
    -------
    np.ndarray
        gradient of the prior neg-log pdf
    """
    g = np.zeros((idx_pix.size, *laplacian_.shape[1:]))  # (n_pix, D)

    for i, n in enumerate(idx_pix):
        mask_i_m = list_edges[:, 1] == n
        mask_i_p = list_edges[:, 0] == n
        neighbors_indices = np.asarray(
            list(list_edges[mask_i_m, 0]) + list(list_edges[mask_i_p, 1]),
            dtype=np.int64,
        )

        g[i] = 2 * (
            neighbors_indices.size * laplacian_[n]
            - np.sum(laplacian_[neighbors_indices, :], axis=0)
        )

    return g  # (n_pix, D)


class L22LaplacianSpatialPrior(SpatialPrior):
    r"""L22 smooth spatial prior, valid for both 1D and 2D tensors. Its pdf is defined as

    .. math::

        \forall d \in [1, D], \quad \pi(\Theta_{\cdot d}) \propto \exp \left[- \tau_d \Vert \Delta \Theta_{\cdot d} \Vert_F^2 \right]

    where  :math:`\Vert \cdot \Vert_F` denotes the Fröbenius norm and :math:`\Delta \Theta_{\cdot d}` is the Laplacian of vector :math:ù\Theta_{\cdot d}`.
    """

    def neglog_pdf(
        self,
        Theta: np.ndarray,
        idx_pix: np.ndarray,
        with_weights: bool = True,
        full: bool = False,
        pixelwise: bool = False,
        chromatic_gibbs: bool = False,
        **kwargs,
    ) -> np.ndarray:
        assert np.sum([pixelwise, full]) < 2
        if idx_pix.size < self.N:
            chromatic_gibbs = True

        chromatic_gibbs = False
        warnings.warn(
            "Warning: Chromatic Gibbs is disabled in the L22LaplacianSpatialPrior"
        )

        if self.list_edges.size > 0:
            if not chromatic_gibbs:
                laplacian_ = compute_laplacian(Theta, self.list_edges, idx_pix=idx_pix)
                laplacian_2_ = laplacian_**2  # (N,D)
                neglog_p = laplacian_2_ * 1
            else:
                laplacian_ = compute_laplacian(
                    Theta, self.list_edges, idx_pix=np.arange(self.N)
                )
                laplacian_2_ = laplacian_**2  # (N,D)
                neglog_p = compute_neglog_chromatic_from_laplacian2(
                    laplacian_2_, list_edges=self.list_edges, idx_pix=idx_pix
                )  # (n_pix, D) or (n_pix, k_mtm, D)

        if with_weights:
            neglog_p *= self.weights

        if full:
            return neglog_p
        elif pixelwise:
            return np.sum(neglog_p, axis=1)
        else:
            return np.sum(neglog_p, axis=0)

    def gradient_neglog_pdf(self, Theta: np.ndarray, idx_pix: np.ndarray) -> np.ndarray:
        assert Theta.shape == (self.N, self.D)

        laplacian_ = compute_laplacian(
            Theta, idx_pix=np.arange(self.N), list_edges=self.list_edges
        )

        g = compute_gradient_from_laplacian(
            laplacian_, idx_pix, list_edges=self.list_edges
        )  # (N, D)
        # g /= self.N * self.D
        return self.weights[None, :] * g  # (N, D)

    def hess_diag_neglog_pdf(
        self, Theta: np.ndarray, idx_pix: np.ndarray
    ) -> np.ndarray:
        hess_diag = np.zeros((idx_pix.size, self.D))  # (n_pix, D)

        if self.list_edges.size > 0:
            idx, counts = np.unique(self.list_edges.flatten(), return_counts=True)
            indices_kept_1 = np.isin(idx_pix, idx, assume_unique=True)
            indices_kept_2 = np.isin(idx, idx_pix, assume_unique=True)

            hess_diag[indices_kept_1, :] += (
                2
                * (counts * (counts + 1))[indices_kept_2, None]
                * np.ones((indices_kept_1.sum(), self.D))
            )

        # hess_diag /= self.N * self.D
        return self.weights[None, :] * hess_diag  # (N, D)
